"""
Monkeypatch the lmfit plotting to avoid TeX errors, and to allow plotting model results in 2D.

This is a very safe monkey patch as we defer to the original plotting function in cases
where it is appropriate, rather than reimplementing this functionality.
"""
import matplotlib.pyplot as plt
from lmfit import model

import xarray as xr

original_plot = model.ModelResult.plot


def transform_lmfit_titles(l, is_title=False):
    if is_title:
        l = l.replace('_', '-')

    return l or ''


def patched_plot(self, *args, **kwargs):
    """
    PyARPES patch for `lmfit` summary plots. Scientists like to have LaTeX in their plots,
    but because underscores outside TeX environments crash matplotlib renders, we need to do
    some fiddling with titles and axis labels in order to prevent users having to switch TeX
    on and off all the time.

    Additionally, this patch provides better support for multidimensional curve fitting.
    :param self:
    :param args:
    :param kwargs:
    :return:
    """
    from arpes.plotting.utils import transform_labels

    try:
        if self.model.n_dims != 1:
            from arpes.plotting.utils import fancy_labels

            fig, ax = plt.subplots(2,2, figsize=(10,8))

            def to_dr(flat_data):
                shape = [len(self.independent[d]) for d in self.independent_order]
                return xr.DataArray(flat_data.reshape(shape), coords=self.independent, dims=self.independent_order)

            to_dr(self.init_fit).plot(ax=ax[1][0])
            to_dr(self.data).plot(ax=ax[0][1])
            to_dr(self.best_fit).plot(ax=ax[0][0])
            to_dr(self.residual).plot(ax=ax[1][1])

            ax[0][0].set_title('Best fit')
            ax[0][1].set_title('Data')
            ax[1][0].set_title('Initial fit')
            ax[1][1].set_title('Residual (Data - Best fit)')

            for axi in ax.ravel():
                fancy_labels(axi)

            plt.tight_layout()
            return ax

    except AttributeError:
        pass

    ret = original_plot(self, *args, **kwargs)
    transform_labels(transform_lmfit_titles)
    return ret


model.ModelResult.plot = patched_plot



import warnings

import numpy as np

from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = (
    'calculate_shirley_background',
    'calculate_shirley_background_full_range',
    'remove_shirley_background',
)

@update_provenance('Remove Shirley background')
def remove_shirley_background(xps: DataType, **kwargs):
    """
    Calculates and removes a Shirley background from a spectrum.
    Only the background corrected spectrum is retrieved.
    :param xps:
    :param kwargs:
    :return:
    """
    xps = normalize_to_spectrum(xps)
    return xps - calculate_shirley_background(xps, **kwargs)


@update_provenance('Calculate full range Shirley background')
def calculate_shirley_background_full_range(xps: DataType, eps=1e-7, max_iters=50, n_samples=5):
    """
    Calculates a shirley background in the range of `energy_slice` according to:

    S(E) = I(E_right) + k * (A_right(E)) / (A_left(E) + A_right(E))

    Typically

    k := I(E_right) - I(E_left)

    The iterative method is continued so long as the total background is not converged to relative error
    `eps`.

    The method continues for a maximum number of iterations `max_iters`.

    In practice, what we can do is to calculate the cumulative sum of the data along the energy axis of
    both the data and the current estimate of the background
    :param xps:
    :param eps:
    :return:
    """

    xps = normalize_to_spectrum(xps)
    background = xps.copy(deep=True)
    cumulative_xps = np.cumsum(xps.values)
    total_xps = np.sum(xps.values)

    rel_error = np.inf

    i_left = np.mean(xps.values[:n_samples])
    i_right = np.mean(xps.values[-n_samples:])

    iter_count = 0

    k = i_left - i_right
    for iter_count in range(max_iters):
        cumulative_background = np.cumsum(background.values)
        total_background = np.sum(background.values)

        new_bkg = background.copy(deep=True)

        for i in range(len(new_bkg)):
            new_bkg.values[i] = i_right + k * (
                (total_xps - cumulative_xps[i] - (total_background - cumulative_background[i])) / (total_xps - total_background + 1e-5)
            )

        rel_error = np.abs(np.sum(new_bkg.values) - total_background) / (total_background)

        background = new_bkg

        if rel_error < eps:
            break

    if (iter_count + 1) == max_iters:
        warnings.warn('Shirley background calculation did not converge ' +
                      'after {} steps with relative error {}!'.format(max_iters, rel_error))

    return background


@update_provenance('Calculate limited range Shirley background')
def calculate_shirley_background(xps: DataType, energy_range: slice=None, eps=1e-7, max_iters=50, n_samples=5):
    """
    Calculates a shirley background iteratively over the full energy range `energy_range`.
    :param xps:
    :param energy_range:
    :param eps:
    :param max_iters:
    :return:
    """
    if energy_range is None:
        energy_range = slice(None, None)

    xps = normalize_to_spectrum(xps)
    xps_for_calc = xps.sel(eV=energy_range)

    bkg = calculate_shirley_background_full_range(xps_for_calc, eps, max_iters)
    full_bkg = xps * 0

    left_idx = np.searchsorted(full_bkg.eV.values, bkg.eV.values[0], side='left')
    right_idx = left_idx + len(bkg)

    full_bkg.values[:left_idx] = bkg.values[0]
    full_bkg.values[left_idx:right_idx] = bkg.values
    full_bkg.values[right_idx:] = bkg.values[-1]

    return full_bkg



import copy
import warnings

import numpy as np

import colorcet as cc
from arpes.fits import ExponentialDecayCModel, GStepBModel
from arpes.io import save_dataset

from .interactive_utils import CursorTool, SaveableTool

__all__ = ('ImageTool',)

# TODO Implement alignment tool

class ImageTool(SaveableTool, CursorTool):
    def __init__(self, curs=None, **kwargs):
        super().__init__(name=kwargs.pop('name', None))

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get('main_width', 600)
        self.app_marginal_size = self.settings.get('marginal_width', 300)

        if curs is not None:
            self.cursor_default = curs

    # TODO select path in image
    def prep_image(self, image_arr):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if self.app_context['color_mode'] == 'linear':
                return image_arr.values

            # avoid dependency conflict with numpy v0.16 for now
            from skimage import exposure  # pylint: disable=import-error
            return exposure.equalize_adapthist(image_arr.values, clip_limit=0.03)

    def tool_handler(self, doc):
        if len(self.arr.shape) == 3:
            return self.tool_handler_3d(doc)

        return self.tool_handler_2d(doc)

    def tool_handler_2d(self, doc):
        from bokeh import events
        from bokeh.layouts import row, column, widgetbox, Spacer
        from bokeh.models import ColumnDataSource, widgets
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models.widgets.markups import Div
        from bokeh.plotting import figure

        arr = self.arr
        # Set up the data
        x_coords, y_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]]

        # Styling
        default_palette = self.default_palette
        if arr.S.is_subtracted:
            default_palette = cc.coolwarm

        error_alpha = 0.3
        error_fill = '#3288bd'

        # Application Organization
        self.app_context.update({
            'data': arr,
            'data_range': {
                'x': (np.min(x_coords.values), np.max(x_coords.values)),
                'y': (np.min(y_coords.values), np.max(y_coords.values)),
            },
            'show_stat_variation': False,
            'color_mode': 'linear',
        })

        def stats_patch_from_data(data, subsampling_rate=None):
            if subsampling_rate is None:
                subsampling_rate = int(min(data.values.shape[0] / 50, 5))
                if subsampling_rate == 0:
                    subsampling_rate = 1

            x_values = data.coords[data.dims[0]].values[::subsampling_rate]
            values = data.values[::subsampling_rate]
            sq = np.sqrt(values)
            lower, upper = values - sq, values + sq

            return {
                'x': np.append(x_values, x_values[::-1]),
                'y': np.append(lower, upper[::-1]),
            }

        def update_stat_variation(plot_name, data):
            patch_data = stats_patch_from_data(data)
            if plot_name != 'right':  # the right plot is on transposed axes
                plots[plot_name + '_marginal_err'].data_source.data = patch_data
            else:
                plots[plot_name + '_marginal_err'].data_source.data = {
                    'x': patch_data['y'],
                    'y': patch_data['x'],
                }

        figures, plots, app_widgets = self.app_context['figures'], self.app_context['plots'], self.app_context[
            'widgets']
        
        if self.cursor_default is not None and len(self.cursor_default) == 2:
            self.cursor = self.cursor_default
        else:
            self.cursor = [np.mean(self.app_context['data_range']['x']),
                           np.mean(self.app_context['data_range']['y'])] # try a sensible default

        # create the main inset plot
        main_image = arr
        prepped_main_image = self.prep_image(main_image)
        self.app_context['color_maps']['main'] = LinearColorMapper(
            default_palette, low=np.min(prepped_main_image), high=np.max(prepped_main_image), nan_color='black')

        main_tools = ["wheel_zoom", "tap", "reset","save"]
        main_title = 'Bokeh Tool: WARNING Unidentified'
        try:
            main_title = "Bokeh Tool: %s" % arr.S.label[:60]
        except:
            pass
        figures['main'] = figure(
            tools=main_tools, plot_width=self.app_main_size, plot_height=self.app_main_size, min_border=10,
            min_border_left=50,
            toolbar_location='left', x_axis_location='below', y_axis_location='right',
            title=main_title, x_range=self.app_context['data_range']['x'],
            y_range=self.app_context['data_range']['y'])
        figures['main'].xaxis.axis_label = arr.dims[0]
        figures['main'].yaxis.axis_label = arr.dims[1]
        figures['main'].toolbar.logo = None
        figures['main'].background_fill_color = "#fafafa"
        plots['main'] = figures['main'].image(
            [prepped_main_image.T], x=self.app_context['data_range']['x'][0], y=self.app_context['data_range']['y'][0],
            dw=self.app_context['data_range']['x'][1] - self.app_context['data_range']['x'][0],
            dh=self.app_context['data_range']['y'][1] - self.app_context['data_range']['y'][0],
            color_mapper=self.app_context['color_maps']['main'])

        app_widgets['info_div'] = Div(text='', width=self.app_marginal_size, height=100)

        # Create the bottom marginal plot
        bottom_marginal = arr.sel(**dict([[arr.dims[1], self.cursor[1]]]), method='nearest')
        figures['bottom_marginal'] = figure(plot_width=self.app_main_size, plot_height=200,
                                            title=None, x_range=figures['main'].x_range,
                                            y_range=(
                                                np.min(bottom_marginal.values), np.max(bottom_marginal.values)),
                                            x_axis_location='above', toolbar_location=None, tools=[])
        plots['bottom_marginal'] = figures['bottom_marginal'].line(
            x=bottom_marginal.coords[arr.dims[0]].values, y=bottom_marginal.values)
        plots['bottom_marginal_err'] = figures['bottom_marginal'].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None)

        # Create the right marginal plot
        right_marginal = arr.sel(**dict([[arr.dims[0], self.cursor[0]]]), method='nearest')
        figures['right_marginal'] = figure(plot_width=200, plot_height=self.app_main_size,
                                           title=None, y_range=figures['main'].y_range,
                                           x_range=(np.min(right_marginal.values), np.max(right_marginal.values)),
                                           y_axis_location='left', toolbar_location=None, tools=[])
        plots['right_marginal'] = figures['right_marginal'].line(
            y=right_marginal.coords[arr.dims[1]].values, x=right_marginal.values)
        plots['right_marginal_err'] = figures['right_marginal'].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None)

        cursor_lines = self.add_cursor_lines(figures['main'])

        # Attach tools and callbacks
        toggle = widgets.Toggle(label="Show Stat. Variation", button_type="success", active=False)

        def set_show_stat_variation(should_show):
            self.app_context['show_stat_variation'] = should_show

            if should_show:
                main_image_data = arr
                update_stat_variation('bottom',
                                      main_image_data.sel(**dict([[arr.dims[1], self.cursor[1]]]), method='nearest'))
                update_stat_variation('right',
                                      main_image_data.sel(**dict([[arr.dims[0], self.cursor[0]]]), method='nearest'))
                plots['bottom_marginal_err'].visible = True
                plots['right_marginal_err'].visible = True
            else:
                plots['bottom_marginal_err'].visible = False
                plots['right_marginal_err'].visible = False

        toggle.on_click(set_show_stat_variation)

        scan_keys = ['x', 'y', 'z', 'pass_energy', 'hv', 'location', 'id',
                     'probe_pol', 'pump_pol']
        scan_info_source = ColumnDataSource({
            'keys': [k for k in scan_keys if k in arr.attrs],
            'values': [str(v) if isinstance(v, float) and np.isnan(v) else v for v in
                       [arr.attrs[k] for k in scan_keys if k in arr.attrs]],
        })
        scan_info_columns = [
            widgets.TableColumn(field='keys', title='Attr.'),
            widgets.TableColumn(field='values', title='Value'),
        ]

        POINTER_MODES = [
            ('Cursor', 'cursor',),
            ('Path', 'path',),
        ]

        COLOR_MODES = [
            ('Adaptive Hist. Eq. (Slow)', 'adaptive_equalization',),
            # ('Histogram Eq.', 'equalization',), # not implemented
            ('Linear', 'linear',),
            # ('Log', 'log',), # not implemented
        ]

        def on_change_color_mode(attr, old, new_color_mode):
            self.app_context['color_mode'] = new_color_mode
            if old is None or old != new_color_mode:
                right_image_data = arr.sel(**dict([[arr.dims[0], self.cursor[0]]]), method='nearest')
                bottom_image_data = arr.sel(**dict([[arr.dims[1], self.cursor[1]]]), method='nearest')
                main_image_data = arr
                prepped_right_image = self.prep_image(right_image_data)
                prepped_bottom_image = self.prep_image(bottom_image_data)
                prepped_main_image = self.prep_image(main_image_data)
                plots['right'].data_source.data = {'image': [prepped_right_image]}
                plots['bottom'].data_source.data = {'image': [prepped_bottom_image.T]}
                plots['main'].data_source.data = {'image': [prepped_main_image.T]}
                update_main_colormap(None, None, main_color_range_slider.value)

        color_mode_dropdown = widgets.Dropdown(label='Color Mode', button_type='primary', menu=COLOR_MODES)
        color_mode_dropdown.on_change('value', on_change_color_mode)

        symmetry_point_name_input = widgets.TextInput(title='Symmetry Point Name', value="G")
        snap_checkbox = widgets.CheckboxButtonGroup(labels=['Snap Axes'], active=[])
        place_symmetry_point_at_cursor_button = widgets.Button(label="Place Point", button_type="primary")

        def update_symmetry_points_for_display():
            pass

        def place_symmetry_point():
            cursor_dict = dict(zip(arr.dims, self.cursor))
            skip_dimensions = {'eV', 'delay', 'cycle'}
            if 'symmetry_points' not in arr.attrs:
                arr.attrs['symmetry_points'] = {}

            snap_distance = {
                'phi': 2,
                'beta': 2,
                'kx': 0.01,
                'ky': 0.01,
                'kz': 0.01,
                'kp': 0.01,
                'hv': 4,
            }

            cursor_dict = {k: v for k, v in cursor_dict.items() if k not in skip_dimensions}
            snapped = copy.copy(cursor_dict)

            if 'Snap Axes' in [snap_checkbox.labels[i] for i in snap_checkbox.active]:
                for axis, value in cursor_dict.items():
                    options = [point[axis] for point in arr.attrs['symmetry_points'].values() if axis in point]
                    options = sorted(options, key=lambda x: np.abs(x - value))
                    if options and np.abs(options[0] - value) < snap_distance[axis]:
                        snapped[axis] = options[0]

            arr.attrs['symmetry_points'][symmetry_point_name_input.value] = snapped

        place_symmetry_point_at_cursor_button.on_click(place_symmetry_point)

        main_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range (Main)')

        layout = row(column(figures['main'], figures['bottom_marginal']),
                     column(figures['right_marginal'], Spacer(width=200, height=200)),
                     column(widgetbox(
                         widgets.Dropdown(label='Pointer Mode', button_type='primary', menu=POINTER_MODES)),
                         widgets.Tabs(tabs=[
                             widgets.Panel(child=widgetbox(
                                 Div(text='<h2>Colorscale:</h2>'),
                                 color_mode_dropdown,
                                 main_color_range_slider,
                                 Div(text='<h2 style="padding-top: 30px;">General Settings:</h2>'),
                                 toggle,
                                 self._cursor_info,
                                 sizing_mode='scale_width'
                             ), title='Settings'),
                             widgets.Panel(child=widgetbox(
                                 app_widgets['info_div'],
                                 Div(text='<h2 style="padding-top: 30px; padding-bottom: 10px;">Scan Info</h2>'),
                                 widgets.DataTable(source=scan_info_source, columns=scan_info_columns, width=400,
                                                   height=400),
                                 sizing_mode='scale_width', width=400
                             ), title='Info'),
                             widgets.Panel(child=widgetbox(
                                 Div(text='<h2>Preparation</h2>'),
                                 symmetry_point_name_input,
                                 snap_checkbox,
                                 place_symmetry_point_at_cursor_button,
                                 sizing_mode='scale_width'
                             ), title='Preparation'),
                         ], width=400)))

        update_main_colormap = self.update_colormap_for('main')

        def on_click_save(event):
            save_dataset(arr)
            print(event)

        def click_main_image(event):
            self.cursor = [event.x, event.y]

            right_marginal_data = arr.sel(**dict([[arr.dims[0], self.cursor[0]]]), method='nearest')
            bottom_marginal_data = arr.sel(**dict([[arr.dims[1], self.cursor[1]]]), method='nearest')
            plots['bottom_marginal'].data_source.data = {
                'x': bottom_marginal_data.coords[arr.dims[0]].values,
                'y': bottom_marginal_data.values,
            }
            plots['right_marginal'].data_source.data = {
                'y': right_marginal_data.coords[arr.dims[1]].values,
                'x': right_marginal_data.values,
            }
            if self.app_context['show_stat_variation']:
                update_stat_variation('right', right_marginal_data)
                update_stat_variation('bottom', bottom_marginal_data)
            figures['bottom_marginal'].y_range.start = np.min(bottom_marginal_data.values)
            figures['bottom_marginal'].y_range.end = np.max(bottom_marginal_data.values)
            figures['right_marginal'].x_range.start = np.min(right_marginal_data.values)
            figures['right_marginal'].x_range.end = np.max(right_marginal_data.values)

            self.save_app()

        figures['main'].on_event(events.Tap, click_main_image)
        main_color_range_slider.on_change('value', update_main_colormap)

        doc.add_root(layout)
        doc.title = "Bokeh Tool"
        self.load_app()
        self.save_app()

    def serialize(self):
        return {
            'cursor_dict': self.cursor_dict,
            'cursor': self.cursor,
        }

    def deserialize(self, json_data):
        if 'cursor' in json_data:
            self.cursor = json_data['cursor']

    def tool_handler_3d(self, doc):
        from bokeh import events
        from bokeh.layouts import row, column, widgetbox, Spacer
        from bokeh.models import ColumnDataSource, HoverTool, widgets
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models.widgets.markups import Div
        from bokeh.plotting import figure

        arr = self.arr
        # Set up the data
        x_coords, y_coords, z_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]], arr.coords[arr.dims[2]]

        info_formatters = {
            'eV': """<div>
                            <h2>Fermi Edge Info:</h2>
                            <p>Gap: <b>{:.1f} meV</b></p>
                            <p>Edge Width: <b>{:.1f} meV</b></p>
                            </div>
                            """,
            'delay': """<div>
                            <h2>Delay Scan Info:</h2>
                            <p>t0: <b>{:.1f} fs</b></p>
                            <p>Decay time: <b>{:.1f} fs</b></p>
                            </div>
                            """
        }

        # Styling
        default_palette = self.default_palette
        if arr.S.is_subtracted:
            default_palette = cc.coolwarm
        error_alpha = 0.3
        error_fill = '#3288bd'

        # Application Organization
        self.app_context.update({
            'data': arr,
            'data_range': {
                'x': (np.min(x_coords.values), np.max(x_coords.values)),
                'y': (np.min(y_coords.values), np.max(y_coords.values)),
                'z': (np.min(z_coords.values), np.max(z_coords.values)),
            },
            'show_stat_variation': False,
            'color_mode': 'linear',
        })

        def stats_patch_from_data(data, subsampling_rate=None):
            if subsampling_rate is None:
                subsampling_rate = int(min(data.values.shape[0] / 50, 5))
                if subsampling_rate == 0:
                    subsampling_rate = 1

            x_values = data.coords[data.dims[0]].values[::subsampling_rate]
            values = data.values[::subsampling_rate]
            sq = np.sqrt(values)
            lower, upper = values - sq, values + sq

            return {
                'x': np.append(x_values, x_values[::-1]),
                'y': np.append(lower, upper[::-1]),
            }

        def update_stat_variation(plot_name, data):
            patch_data = stats_patch_from_data(data)
            if plot_name != 'right':  # the right plot is on transposed axes
                plots[plot_name + '_marginal_err'].data_source.data = patch_data
            else:
                plots[plot_name + '_marginal_err'].data_source.data = {
                    'x': patch_data['y'],
                    'y': patch_data['x'],
                }

        figures, plots, app_widgets = self.app_context['figures'], self.app_context['plots'], self.app_context[
            'widgets']
        
        if self.cursor_default is not None and len(self.cursor_default) == 3:
            self.cursor = self.cursor_default
        else:
            self.cursor = [np.mean(self.data_range['x']),
                           np.mean(self.data_range['y']),
                           np.mean(self.data_range['z'])] # try a sensible default

        # create the main inset plot
        main_image = arr.sel(**dict([[arr.dims[2], self.cursor[2]]]), method='nearest')
        prepped_main_image = self.prep_image(main_image)
        self.app_context['color_maps']['main'] = LinearColorMapper(
            default_palette, low=np.min(prepped_main_image), high=np.max(prepped_main_image), nan_color='black')

        main_title = 'Bokeh Tool: WARNING Unidentified'
        try:
            main_title = "Bokeh Tool: %s" % arr.S.label[:60]
        except:
            pass

        main_tools = ["wheel_zoom", "tap", "reset","save"]
        figures['main'] = figure(
            tools=main_tools, plot_width=self.app_main_size, plot_height=self.app_main_size, min_border=10,
            min_border_left=50,
            toolbar_location='left', x_axis_location='below', y_axis_location='right',
            title=main_title, x_range=self.app_context['data_range']['x'],
            y_range=self.app_context['data_range']['y'])
        figures['main'].xaxis.axis_label = arr.dims[0]
        figures['main'].yaxis.axis_label = arr.dims[1]
        figures['main'].toolbar.logo = None
        figures['main'].background_fill_color = "#fafafa"
        plots['main'] = figures['main'].image(
            [prepped_main_image.T], x=self.app_context['data_range']['x'][0], y=self.app_context['data_range']['y'][0],
            dw=self.app_context['data_range']['x'][1] - self.app_context['data_range']['x'][0],
            dh=self.app_context['data_range']['y'][1] - self.app_context['data_range']['y'][0],
            color_mapper=self.app_context['color_maps']['main'])

        # Create the z-selector
        z_marginal_data = arr.sel(
            **dict([[arr.dims[0], self.cursor[0]], [arr.dims[1], self.cursor[1]]]),
            method='nearest')
        z_hover_tool = HoverTool(
            tooltips=[
                ('x', '@x{1.111}'),
                ('Int. (arb.)', '@y{1.1}'),
            ],
            mode='vline'
        )
        z_tools = [z_hover_tool, 'wheel_zoom', 'tap', 'reset']
        figures['z_marginal'] = figure(
            plot_width=self.app_marginal_size, plot_height=self.app_marginal_size,
            x_range=self.app_context['data_range']['z'],
            y_range=(np.min(z_marginal_data.values), np.max(z_marginal_data.values)),
            x_axis_location='above', toolbar_location='below', y_axis_location='right', tools=z_tools)
        figures['z_marginal'].xaxis.major_label_text_font_size = '0pt'
        figures['z_marginal'].xaxis.axis_label = arr.dims[2]
        figures['z_marginal'].toolbar.logo = None
        plots['z_marginal'] = figures['z_marginal'].line(x=z_coords.values, y=z_marginal_data.values)
        plots['z_marginal_err'] = figures['z_marginal'].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None)

        info_formatter = info_formatters.get(arr.dims[2], '')
        app_widgets['info_div'] = Div(text='', width=self.app_marginal_size, height=100)

        if arr.dims[2] == 'eV':
            # Try to fit a Fermi edge and display it in the plot
            try:
                fit_data = z_marginal_data.sel(eV=slice(-0.25, 0.2))
                z_fit = GStepBModel().guess_fit(fit_data)
                plots['z_fit'] = figures['z_marginal'].line(x=fit_data.coords['eV'].values, y=z_fit.best_fit,
                                                        line_dash='dashed', line_color='red')
                self.app_context['z_model'] = GStepBModel

                app_widgets['info_div'].text = info_formatter.format(
                    z_fit.params['center'].value * 1000,
                    z_fit.params['width'].value * 1000,
                )
            except:
                pass

        elif arr.dims[2] == 'delay' and 't0' in arr.attrs:
            # Try to fit a decay constant to the data after t0
            plots['t0_marker'] = figures['z_marginal'].line(
                x=[float(arr.attrs.get('t0')), float(arr.attrs.get('t0'))], y=[0, 1000000], line_color='black',
                line_dash='dashed')
            t0 = float(arr.attrs['t0'])
            self.app_context['z_model'] = ExponentialDecayCModel
            try:
                after_t0 = z_marginal_data.sel(delay=slice(t0 - 0.2, None))
                exp_model = ExponentialDecayCModel()
                z_fit = exp_model.guess_fit(after_t0, params={'t0': {'value': t0}})
                plots['z_fit'] = figures['z_marginal'].line(
                    x=after_t0.coords['delay'].values, y=z_fit.best_fit, line_dash='dashed', line_color='red')
            except Exception as e:
                plots['z_fit'] = figures['z_marginal'].line(x=[], y=[], line_dash='dashed', line_color='red')

        # Create the bottom marginal plot
        bottom_image = arr.sel(**dict([[arr.dims[1], self.cursor[1]]]), method='nearest')
        prepped_bottom_image = self.prep_image(bottom_image)
        self.app_context['color_maps']['bottom'] = LinearColorMapper(
            default_palette, low=np.min(prepped_bottom_image), high=np.max(prepped_bottom_image), nan_color='black')
        figures['bottom'] = figure(plot_width=self.app_main_size, plot_height=self.app_marginal_size,
                                   title=None, x_range=figures['main'].x_range,
                                   y_range=figures['z_marginal'].x_range,
                                   x_axis_location='above',
                                   toolbar_location=None, tools=[])
        figures['bottom'].xaxis.major_label_text_font_size = '0pt'
        plots['bottom'] = figures['bottom'].image([prepped_bottom_image.T], x=self.app_context['data_range']['x'][0],
                                                  y=self.app_context['data_range']['z'][0],
                                                  dw=self.app_context['data_range']['x'][1] -
                                                     self.app_context['data_range']['x'][0],
                                                  dh=self.app_context['data_range']['z'][1] -
                                                     self.app_context['data_range']['z'][0],
                                                  color_mapper=self.app_context['color_maps']['bottom'])
        bottom_marginal = bottom_image.sel(**dict([[arr.dims[2], self.cursor[2]]]), method='nearest')
        figures['bottom_marginal'] = figure(plot_width=self.app_main_size, plot_height=200,
                                            title=None, x_range=figures['main'].x_range,
                                            y_range=(
                                                np.min(bottom_marginal.values), np.max(bottom_marginal.values)),
                                            x_axis_location='above', toolbar_location=None, tools=[])
        plots['bottom_marginal'] = figures['bottom_marginal'].line(
            x=bottom_marginal.coords[arr.dims[0]].values, y=bottom_marginal.values)
        plots['bottom_marginal_err'] = figures['bottom_marginal'].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None)

        # Create the right marginal plot
        right_image = arr.sel(**dict([[arr.dims[0], self.cursor[0]]]), method='nearest')
        prepped_right_image = self.prep_image(right_image)
        self.app_context['color_maps']['right'] = LinearColorMapper(
            default_palette, low=np.min(prepped_right_image), high=np.max(prepped_right_image), nan_color='black')
        figures['right'] = figure(plot_width=self.app_marginal_size, plot_height=self.app_main_size, title=None,
                                  x_range=figures['z_marginal'].x_range, y_range=figures['main'].y_range,
                                  toolbar_location=None, tools=[])
        figures['right'].yaxis.major_label_text_font_size = '0pt'
        plots['right'] = figures['right'].image([prepped_right_image], x=self.app_context['data_range']['z'][0],
                                                y=self.app_context['data_range']['y'][0],
                                                dw=self.app_context['data_range']['z'][1] -
                                                   self.app_context['data_range']['z'][0],
                                                dh=self.app_context['data_range']['y'][1] -
                                                   self.app_context['data_range']['y'][0],
                                                color_mapper=self.app_context['color_maps']['right'])
        right_marginal = right_image.sel(**dict([[arr.dims[2], self.cursor[2]]]), method='nearest')
        figures['right_marginal'] = figure(plot_width=200, plot_height=self.app_main_size,
                                           title=None, y_range=figures['main'].y_range,
                                           x_range=(np.min(right_marginal.values), np.max(right_marginal.values)),
                                           y_axis_location='left', toolbar_location=None, tools=[])
        plots['right_marginal'] = figures['right_marginal'].line(
            y=right_marginal.coords[arr.dims[1]].values, x=right_marginal.values)
        plots['right_marginal_err'] = figures['right_marginal'].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None)

        cursor_lines = self.add_cursor_lines(figures['main'])

        # Attach tools and callbacks
        toggle = widgets.Toggle(label="Show Stat. Variation", button_type="success", active=False)

        def set_show_stat_variation(should_show):
            self.app_context['show_stat_variation'] = should_show

            if should_show:
                main_image_data = arr.sel(**dict([[arr.dims[2], self.cursor[2]]]), method='nearest')
                update_stat_variation('z', arr.sel(**dict([[arr.dims[0], self.cursor[0]], [arr.dims[1], self.cursor[1]]]),
                                                   method='nearest'))
                update_stat_variation('bottom',
                                      main_image_data.sel(**dict([[arr.dims[1], self.cursor[1]]]), method='nearest'))
                update_stat_variation('right',
                                      main_image_data.sel(**dict([[arr.dims[0], self.cursor[0]]]), method='nearest'))
                plots['z_marginal_err'].visible = True
                plots['bottom_marginal_err'].visible = True
                plots['right_marginal_err'].visible = True
            else:
                plots['z_marginal_err'].visible = False
                plots['bottom_marginal_err'].visible = False
                plots['right_marginal_err'].visible = False

        toggle.on_click(set_show_stat_variation)

        scan_keys = ['x', 'y', 'z', 'pass_energy', 'hv', 'location', 'id',
                     'probe_pol', 'pump_pol']
        scan_info_source = ColumnDataSource({
            'keys': [k for k in scan_keys if k in arr.attrs],
            'values': [str(v) if isinstance(v, float) and np.isnan(v) else v for v in
                       [arr.attrs[k] for k in scan_keys if k in arr.attrs]],
        })
        scan_info_columns = [
            widgets.TableColumn(field='keys', title='Attr.'),
            widgets.TableColumn(field='values', title='Value'),
        ]

        POINTER_MODES = [
            ('Cursor', 'cursor',),
            ('Path', 'path',),
        ]

        COLOR_MODES = [
            ('Adaptive Hist. Eq. (Slow)', 'adaptive_equalization',),
            # ('Histogram Eq.', 'equalization',), # not implemented
            ('Linear', 'linear',),
            # ('Log', 'log',), # not implemented
        ]

        def on_change_color_mode(attr, old, new_color_mode):
            self.app_context['color_mode'] = new_color_mode
            if old is None or old != new_color_mode:
                cursor = self.cursor
                right_image_data = arr.sel(**dict([[arr.dims[0], cursor[0]]]), method='nearest')
                bottom_image_data = arr.sel(**dict([[arr.dims[1], cursor[1]]]), method='nearest')
                main_image_data = arr.sel(**dict([[arr.dims[2], cursor[2]]]), method='nearest')
                prepped_right_image = self.prep_image(right_image_data)
                prepped_bottom_image = self.prep_image(bottom_image_data)
                prepped_main_image = self.prep_image(main_image_data)
                plots['right'].data_source.data = {'image': [prepped_right_image]}
                plots['bottom'].data_source.data = {'image': [prepped_bottom_image.T]}
                plots['main'].data_source.data = {'image': [prepped_main_image.T]}
                update_right_colormap(None, None, right_color_range_slider.value)
                update_bottom_colormap(None, None, bottom_color_range_slider.value)
                update_main_colormap(None, None, main_color_range_slider.value)

        color_mode_dropdown = widgets.Dropdown(label='Color Mode', button_type='primary', menu=COLOR_MODES)
        color_mode_dropdown.on_change('value', on_change_color_mode)

        symmetry_point_name_input = widgets.TextInput(title='Symmetry Point Name', value="G")
        snap_checkbox = widgets.CheckboxButtonGroup(labels=['Snap Axes'], active=[])
        place_symmetry_point_at_cursor_button = widgets.Button(label="Place Point", button_type="primary")

        def update_symmetry_points_for_display():
            pass

        def place_symmetry_point():
            cursor_dict = dict(zip(arr.dims, self.cursor))
            skip_dimensions = {'eV', 'delay', 'cycle'}
            if 'symmetry_points' not in arr.attrs:
                arr.attrs['symmetry_points'] = {}

            snap_distance = {
                'phi': 2,
                'beta': 2,
                'kx': 0.01,
                'ky': 0.01,
                'kz': 0.01,
                'kp': 0.01,
                'hv': 4,
            }

            cursor_dict = {k: v for k, v in cursor_dict.items() if k not in skip_dimensions}
            snapped = copy.copy(cursor_dict)

            if 'Snap Axes' in [snap_checkbox.labels[i] for i in snap_checkbox.active]:
                for axis, value in cursor_dict.items():
                    options = [point[axis] for point in arr.attrs['symmetry_points'].values() if axis in point]
                    options = sorted(options, key=lambda x: np.abs(x - value))
                    if options and np.abs(options[0] - value) < snap_distance[axis]:
                        snapped[axis] = options[0]

            arr.attrs['symmetry_points'][symmetry_point_name_input.value] = snapped

        place_symmetry_point_at_cursor_button.on_click(place_symmetry_point)

        main_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range (Main)')
        right_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range (%s Marginal)' % arr.dims[1])
        bottom_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range (%s Marginal)' % arr.dims[0])
        layout = row(column(figures['main'], figures['bottom'], figures['bottom_marginal']),
                     column(figures['right'], figures['z_marginal'], Spacer(width=self.app_marginal_size, height=200)),
                     column(figures['right_marginal'], Spacer(width=200, height=self.app_marginal_size),
                            Spacer(width=200, height=200)),
                     column(widgetbox(
                         widgets.Dropdown(label='Pointer Mode', button_type='primary', menu=POINTER_MODES)),
                         widgets.Tabs(tabs=[
                             widgets.Panel(child=widgetbox(
                                 Div(text='<h2>Colorscale:</h2>'),
                                 color_mode_dropdown,
                                 main_color_range_slider,
                                 right_color_range_slider,
                                 bottom_color_range_slider,
                                 Div(text='<h2 style="padding-top: 30px;">General Settings:</h2>'),
                                 toggle,
                                 self._cursor_info,
                                 sizing_mode='scale_width'
                             ), title='Settings'),
                             widgets.Panel(child=widgetbox(
                                 app_widgets['info_div'],
                                 Div(text='<h2 style="padding-top: 30px; padding-bottom: 10px;">Scan Info</h2>'),
                                 widgets.DataTable(source=scan_info_source, columns=scan_info_columns, width=400,
                                                   height=400),
                                 sizing_mode='scale_width', width=400
                             ), title='Info'),
                             widgets.Panel(child=widgetbox(
                                 Div(text='<h2>Preparation</h2>'),
                                 symmetry_point_name_input,
                                 snap_checkbox,
                                 place_symmetry_point_at_cursor_button,
                                 sizing_mode='scale_width'
                             ), title='Preparation'),
                         ], width=400)))

        update_main_colormap = self.update_colormap_for('main')
        update_bottom_colormap = self.update_colormap_for('bottom')
        update_right_colormap = self.update_colormap_for('right')

        def on_click_save(event):
            save_dataset(arr)
            print(event)

        def click_z_marginal(event):
            self.cursor = [self.cursor[0], self.cursor[1], event.x]
            cursor = self.cursor

            main_image = arr.sel(**dict([[arr.dims[2], cursor[2]]]), method='nearest')
            plots['main'].data_source.data = {
                'image': [self.prep_image(main_image).T]
            }
            update_main_colormap(None, None, main_color_range_slider.value)
            right_marginal_data = main_image.sel(**dict([[arr.dims[0], cursor[0]]]), method='nearest')
            bottom_marginal_data = main_image.sel(**dict([[arr.dims[1], cursor[1]]]), method='nearest')
            plots['bottom_marginal'].data_source.data = {
                'x': bottom_marginal_data.coords[arr.dims[0]].values,
                'y': bottom_marginal_data.values,
            }
            plots['right_marginal'].data_source.data = {
                'y': right_marginal_data.coords[arr.dims[1]].values,
                'x': right_marginal_data.values,
            }
            if self.app_context['show_stat_variation']:
                update_stat_variation('right', right_marginal_data)
                update_stat_variation('bottom', bottom_marginal_data)
            figures['bottom_marginal'].y_range.start = np.min(bottom_marginal_data.values)
            figures['bottom_marginal'].y_range.end = np.max(bottom_marginal_data.values)
            figures['right_marginal'].x_range.start = np.min(right_marginal_data.values)
            figures['right_marginal'].x_range.end = np.max(right_marginal_data.values)

        def click_main_image(event):
            self.cursor = [event.x, event.y, self.cursor[2]]
            cursor = self.cursor
            right_image_data = arr.sel(**dict([[arr.dims[0], cursor[0]]]), method='nearest')
            bottom_image_data = arr.sel(**dict([[arr.dims[1], cursor[1]]]), method='nearest')
            prepped_right_image = self.prep_image(right_image_data)
            prepped_bottom_image = self.prep_image(bottom_image_data)
            plots['right'].data_source.data = {'image': [prepped_right_image]}
            plots['bottom'].data_source.data = {'image': [prepped_bottom_image.T]}
            update_right_colormap(None, None, right_color_range_slider.value)
            update_bottom_colormap(None, None, bottom_color_range_slider.value)
            right_marginal_data = right_image_data.sel(**dict([[arr.dims[2], cursor[2]]]), method='nearest')
            bottom_marginal_data = bottom_image_data.sel(**dict([[arr.dims[2], cursor[2]]]), method='nearest')
            z_data = arr.sel(**dict([[arr.dims[0], cursor[0]], [arr.dims[1], cursor[1]]]), method='nearest')
            plots['z_marginal'].data_source.data = {
                'x': z_coords.values,
                'y': z_data.values,
            }
            plots['bottom_marginal'].data_source.data = {
                'x': bottom_marginal_data.coords[arr.dims[0]].values,
                'y': bottom_marginal_data.values,
            }
            plots['right_marginal'].data_source.data = {
                'y': right_marginal_data.coords[arr.dims[1]].values,
                'x': right_marginal_data.values,
            }
            if self.app_context['show_stat_variation']:
                update_stat_variation('z', z_data)
                update_stat_variation('right', right_marginal_data)
                update_stat_variation('bottom', bottom_marginal_data)
            figures['z_marginal'].y_range.start = np.min(z_data.values)
            figures['z_marginal'].y_range.end = np.max(z_data.values)
            figures['bottom_marginal'].y_range.start = np.min(bottom_marginal_data.values)
            figures['bottom_marginal'].y_range.end = np.max(bottom_marginal_data.values)
            figures['right_marginal'].x_range.start = np.min(right_marginal_data.values)
            figures['right_marginal'].x_range.end = np.max(right_marginal_data.values)
            if 'z_fit' in plots and arr.dims[2] == 'eV':
                z_fit_data = z_data.sel(eV=slice(-0.25, 0.2))
                new_z_fit = GStepBModel().guess_fit(z_fit_data)
                plots['z_fit'].data_source.data = {
                    'x': z_fit_data.coords['eV'].values,
                    'y': new_z_fit.best_fit,
                }
                app_widgets['info_div'].text = info_formatter.format(
                    new_z_fit.params['center'].value * 1000,
                    new_z_fit.params['width'].value * 1000,
                )
            if 'z_fit' in plots and arr.dims[2] == 'delay':
                try:
                    t0 = float(arr.attrs.get('t0', 0))
                    after_t0 = z_data.sel(delay=slice(t0 - 0.2, None))
                    exp_model = ExponentialDecayCModel()
                    z_fit = exp_model.guess_fit(after_t0, params={'t0': {'value': t0}})
                    plots['z_fit'].data_source.data = {
                        'x': after_t0.coords['delay'],
                        'y': z_fit.best_fit,
                    }
                except Exception as e:
                    plots['z_fit'].data_source.data = {
                        'x': [],
                        'y': [],
                    }

            self.save_app()

        figures['z_marginal'].on_event(events.Tap, click_z_marginal)
        figures['main'].on_event(events.Tap, click_main_image)
        main_color_range_slider.on_change('value', update_main_colormap)
        bottom_color_range_slider.on_change('value', update_bottom_colormap)
        right_color_range_slider.on_change('value', update_right_colormap)

        doc.add_root(layout)
        doc.title = "Bokeh Tool"
        self.load_app()
        self.save_app()



import matplotlib.cm as cm
import matplotlib.colors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import LineCollection
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import xarray as xr
from arpes.analysis.sarpes import to_intensity_polarization
from arpes.analysis.statistics import mean_and_deviation
from arpes.bootstrap import bootstrap
from arpes.plotting.tof import scatter_with_std
from arpes.plotting.utils import label_for_dim, savefig, path_for_plot, polarization_colorbar
from arpes.provenance import save_plot_provenance
from arpes.utilities.math import polarization, propagate_statistical_error


__all__ = ('spin_polarized_spectrum', 'spin_colored_spectrum',
           'spin_difference_spectrum',)


test_polarization = propagate_statistical_error(polarization)

@save_plot_provenance
def spin_colored_spectrum(spin_dr, title=None, ax=None, out=None, scatter=False, **kwargs):
    if ax is None:
        _, ax = plt.subplots(figsize=(6, 4))

    as_intensity = to_intensity_polarization(spin_dr)
    intensity = as_intensity.intensity
    pol = as_intensity.polarization.copy(deep=True)

    if len(intensity.dims) == 1:
        inset_ax = inset_axes(ax, width="30%", height="5%", loc=1)
        coord = intensity.coords[intensity.dims[0]]
        points = np.array([coord.values, intensity.values]).T.reshape(-1, 1, 2)
        pol.values[np.isnan(pol.values)] = 0
        pol.values[pol.values > 1] = 1
        pol.values[pol.values < -1] = -1
        pol_colors = cm.get_cmap('RdBu')(pol.values[:-1])

        if scatter:
            pol_colors = cm.get_cmap('RdBu')(pol.values)
            ax.scatter(coord.values, intensity.values, c=pol_colors, s=1.5)
        else:
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            lc = LineCollection(segments, colors=pol_colors)

            ax.add_collection(lc)

        ax.set_xlim(coord.min().item(), coord.max().item())
        ax.set_ylim(0, intensity.max().item() * 1.15)
        ax.set_ylabel('ARPES Spectrum Intensity (arb.)')
        ax.set_xlabel(label_for_dim(spin_dr, dim_name=intensity.dims[0]))
        ax.set_title(title if title is not None else 'Spin Polarization')
        polarization_colorbar(inset_ax)


    if out is not None:
        savefig(out, dpi=400)
        plt.clf()
        return path_for_plot(out)
    else:
        plt.show()

@save_plot_provenance
def spin_difference_spectrum(spin_dr, title=None, ax=None, out=None, scatter=False, **kwargs):
    if ax is None:
        _, ax = plt.subplots(figsize=(6, 4))

    try:
        as_intensity = to_intensity_polarization(spin_dr)
    except AssertionError:
        as_intensity = spin_dr
    intensity = as_intensity.intensity
    pol = as_intensity.polarization.copy(deep=True)

    if len(intensity.dims) == 1:
        inset_ax = inset_axes(ax, width="30%", height="5%", loc=1)
        coord = intensity.coords[intensity.dims[0]]
        points = np.array([coord.values, intensity.values]).T.reshape(-1, 1, 2)
        pol.values[np.isnan(pol.values)] = 0
        pol.values[pol.values > 1] = 1
        pol.values[pol.values < -1] = -1
        pol_colors = cm.get_cmap('RdBu')(pol.values[:-1])

        if scatter:
            pol_colors = cm.get_cmap('RdBu')(pol.values)
            ax.scatter(coord.values, intensity.values, c=pol_colors, s=1.5)
        else:
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            lc = LineCollection(segments, colors=pol_colors)

            ax.add_collection(lc)

        ax.set_xlim(coord.min().item(), coord.max().item())
        ax.set_ylim(0, intensity.max().item() * 1.15)
        ax.set_ylabel('ARPES Spectrum Intensity (arb.)')
        ax.set_xlabel(label_for_dim(spin_dr, dim_name=intensity.dims[0]))
        ax.set_title(title if title is not None else 'Spin Polarization')
        polarization_colorbar(inset_ax)

    if out is not None:
        savefig(out, dpi=400)
        plt.clf()
        return path_for_plot(out)
    else:
        plt.show()


@save_plot_provenance
def spin_polarized_spectrum(
        spin_dr, title=None, ax=None, out=None, component='y', scatter=False, stats=False, norm=None):
    if ax is None:
        _, ax = plt.subplots(2, 1, sharex=True)

    if stats:
        spin_dr = bootstrap(lambda x: x)(spin_dr, N=100)
        pol = mean_and_deviation(to_intensity_polarization(spin_dr))
        counts = mean_and_deviation(spin_dr)
    else:
        counts = spin_dr
        pol = to_intensity_polarization(counts)

    ax_left = ax[0]
    ax_right = ax[1]

    up = counts.down.data
    down = counts.up.data

    energies = spin_dr.coords['eV'].values
    min_e, max_e = np.min(energies), np.max(energies)

    # Plot the spectra
    if stats:
        if scatter:
            scatter_with_std(counts, 'up', color='red', ax=ax_left)
            scatter_with_std(counts, 'down', color='blue', ax=ax_left)
        else:
            v, s = counts.up.values, counts.up_std.values
            ax_left.plot(energies, v, 'r')
            ax_left.fill_between(energies, v - s, v + s, color='r', alpha=0.25)

            v, s = counts.down.values, counts.down_std.values
            ax_left.plot(energies, v, 'b')
            ax_left.fill_between(energies, v - s, v + s, color='b', alpha=0.25)
    else:
        ax_left.plot(energies, up, 'r')
        ax_left.plot(energies, down, 'b')

    ax_left.set_title(title if title is not None else 'Spin spectrum {}'.format(''))
    ax_left.set_ylabel(r'\textbf{Spectrum Intensity}')
    ax_left.set_xlabel(r'\textbf{Kinetic energy} (eV)')
    ax_left.set_xlim(min_e, max_e)

    max_up = np.max(up)
    max_down = np.max(down)
    ax_left.set_ylim(0, max(max_down, max_up) * 1.2)

    # Plot the polarization and associated statistical error bars
    if stats:
        if scatter:
            scatter_with_std(pol, 'polarization', ax=ax_right, color='black')
        else:
            v = pol.polarization.data
            s = pol.polarization_std.data
            ax_right.plot(energies, v, color='black')
            ax_right.fill_between(energies, v - s, v + s, color='black', alpha=0.25)

    else:
        ax_right.plot(energies, pol.polarization.data, color='black')
    ax_right.fill_between(energies, 0, 1, facecolor='blue', alpha=0.1)
    ax_right.fill_between(energies, -1, 0, facecolor='red', alpha=0.1)

    ax_right.set_title('Spin polarization, $\\text{S}_\\textbf{' + component + '}$')
    ax_right.set_ylabel(r'\textbf{Polarization}')
    ax_right.set_xlabel(r'\textbf{Kinetic Energy} (eV)')
    ax_right.set_xlim(min_e, max_e)
    ax_right.axhline(0, color='white', linestyle=':')

    ax_right.set_ylim(-1, 1)
    ax_right.grid(True, axis='y')

    plt.tight_layout()

    if out is not None:
        savefig(out, dpi=400)
        plt.clf()
        return path_for_plot(out)
    else:
        pass

    return ax

def polarization_intensity_to_color(data: xr.Dataset, vmax=None, pmax=1):
    """
    Converts a dataset with intensity and polarization into a RGB colorarray. This consists of a few steps:

    1. first we take the polarization to get a RdBu RGB value
    2. We convert the RGB value to HSV
    3. We use the relative intensity to compute a new value for the V ('value') channel
    4. We convert back to RGB
    :param data:
    :return:
    """

    if vmax is None:
        # use the 98th percentile data if not provided
        vmax = np.percentile(data.intensity.values, 98)

    rgbas = cm.RdBu((data.polarization.values/pmax + 1) / 2)
    slices = [slice(None) for _ in data.polarization.dims] + [slice(0, 3)]
    rgbs = rgbas[slices]

    hsvs = matplotlib.colors.rgb_to_hsv(rgbs)

    intensity_values = data.intensity.values.copy() / vmax
    intensity_values[intensity_values > 1] = 1
    hsvs[:, :, 2] = intensity_values

    return matplotlib.colors.hsv_to_rgb(hsvs)


@save_plot_provenance
def hue_brightness_plot(data: xr.Dataset, ax=None, out=None, **kwargs):
    assert('intensity' in data and 'polarization' in data)

    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get('figsize', (7, 5,)))

    x, y = data.coords[data.intensity.dims[0]].values, data.coords[data.intensity.dims[1]].values
    extent = [y[0], y[-1], x[0], x[-1]]
    ax.imshow(polarization_intensity_to_color(data, **kwargs), extent=extent, aspect='auto', origin='lower')
    ax.set_xlabel(data.intensity.dims[1])
    ax.set_ylabel(data.intensity.dims[0])

    ax.grid(False)


    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax



import collections
import itertools

import numpy as np
import scipy.ndimage.interpolation

import arpes.constants
import xarray as xr


def derivative(f, arg_idx=0):
    """
    Defines a simple midpoint derivative
    """

    def d(*args):
        args = list(args)
        ref_arg = args[arg_idx]
        d = ref_arg / 100

        args[arg_idx] = ref_arg + d
        high = f(*args)
        args[arg_idx] = ref_arg - d
        low = f(*args)

        return (high - low) / (2 * d)

    return d


def polarization(up, down):
    return (up - down) / (up + down)


def propagate_statistical_error(f):
    def compute_propagated_error(*args):
        running_sum = 0
        for i, arg in enumerate(args):
            df_darg_i = derivative(f, i)
            running_sum += df_darg_i(*args) ** 2 * arg

        return np.sqrt(running_sum)

    return compute_propagated_error

def shift_by(arr, value, axis=0, by_axis=0, **kwargs):
    assert axis != by_axis
    arr_copy = arr.copy()

    if isinstance(value, xr.DataArray):
        value = value.values

    if not isinstance(value, collections.Iterable):
        value = list(itertools.repeat(value, times=arr.shape[by_axis]))

    for axis_idx in range(arr.shape[by_axis]):
        slc = (slice(None),) * by_axis + (axis_idx,) + (slice(None),) * (arr.ndim - by_axis - 1)
        shift_amount = (0,) * axis + (value[axis_idx],) + (0,) * (arr.ndim - axis - 1)

        if axis > by_axis:
            shift_amount = shift_amount[1:]
        else:
            shift_amount = shift_amount[:-1]

        arr_copy[slc] = scipy.ndimage.interpolation.shift(arr[slc], shift_amount, **kwargs)

    return arr_copy


def inv_fermi_distribution(energy, temperature, mu=0):
    """Expects energy in eV and temperature in Kelvin."""
    return np.exp((energy - mu) / (arpes.constants.K_BOLTZMANN_EV_KELVIN * temperature)) + 1


def fermi_distribution(energy, temperature):
    """Expects energy in eV and temperature in Kelvin."""
    return 1 / inv_fermi_distribution(energy, temperature)



# pylint: disable=unused-import

import warnings
import contextlib

from typing import Union

VERSION = '2.1.3'


def check() -> None:
    """
    Verifies certain aspects of the installation and provides guidance for semi-broken installations.
    :return:
    """

    def verify_qt_tool() -> Union[str, None]:
        pip_command = 'pip install pyqtgraph'
        warning = 'Using qt_tool, the PyARPES version of Image Tool, requires ' \
                  'pyqtgraph and Qt5:\n\n\tYou can install with: {}'.format(pip_command)
        try:
            import pyqtgraph
        except ImportError:
            return warning
        return None

    def verify_igor_pro() -> Union[str, None]:
        pip_command = 'pip install https://github.com/chstan/igorpy/tarball/712a4c4#egg=igor-0.3.1'
        warning = 'For Igor support, install igorpy with: {}'.format(pip_command)
        warning_incompatible = 'PyARPES requires a patched copy of igorpy, ' \
                               'available at \n\thttps://github.com/chstan/igorpy/tarball/712a4c4\n\n\tYou can install with: ' \
                               '{}'.format(pip_command)
        try:
            import igor
            if igor.__version__ <= '0.3':
                raise ValueError('Not using patched version of igorpy.')

        except ValueError:
            return warning_incompatible
        except (ImportError, AttributeError):
            return warning

        return None

    def verify_bokeh() -> Union[str, None]:
        pip_command = 'pip install bokeh==0.12.10'

        warning = 'For bokeh support, install version 0.12.10\n\t with {}'.format(pip_command)
        warning_incompatible = 'PyARPES, requires version 0.12.10 of bokeh. You can install with \n\t{}'.format(pip_command)

        try:
            import bokeh
            if not bokeh.__version__ == '0.12.10':
                raise ValueError('Not using the specified version of Bokeh.')

        except ImportError:
            return warning
        except ValueError:
            return warning_incompatible
        return None

    def verify_everything() -> Union[str, None]:
        import os
        phony_datasets = os.path.abspath(__file__ + '../../../tests/resources')
        try:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                with open(os.devnull, 'w') as devnull:
                    with contextlib.redirect_stdout(devnull):
                        setup({}, arpes_root=phony_datasets)

        except Exception as e: # pylint: disable=broad-except
            return str(e)

        return None

    checks = [
        ('Igor Pro Support', verify_igor_pro,),
        ('Bokeh Support', verify_bokeh,),
        ('qt_tool Support', verify_qt_tool,),
        ('Import almost everything in PyARPES', verify_everything,),
    ]

    from colorama import Fore, Style

    print('Checking...')
    for check_name, check_fn in checks:
        initial_str = '[ ] {}'.format(check_name)
        print(initial_str, end='', flush=True)

        failure_message = check_fn()

        print('\b' * len(initial_str) + ' ' * len(initial_str) + '\b' * len(initial_str), end='')

        if failure_message is None:
            print('{}[✔] {}{}'.format(Fore.GREEN, check_name, Style.RESET_ALL))
        else:
            print('{}[✘] {}: \n\t{}{}'.format(Fore.RED, check_name, failure_message, Style.RESET_ALL))


def setup(outer_globals, arpes_root=None):
    """
    Performs standard imports similar to what might be achieved by specifying a reasonable IPython startup file.

    Consider this a "basic" configuration function for those that do not want to deal with IPython configuration
    themselves.

    As a bonus, this file tracks with the repository so that function moves should happen fairly transparently, and new
    functionality becomes "automatically" available.

    :param outer_globals: Globals for the calling module scope, this should be provided by calling setup(globals(), ...)
    :param arpes_root: Location of the analysis directory you are working in, as generated by
    :return:
    """

    import os

    if arpes_root is None and os.getenv('ARPES_ROOT') is None:
        raise ValueError('You must supply the location of your analysis directory.')
    if arpes_root is not None:
        os.environ['ARPES_ROOT'] = arpes_root

    import arpes.config
    arpes.config.update_configuration(user_path=os.environ['ARPES_ROOT'])

    from importlib import import_module

    def global_import_from(module_name, attributes):
        if not isinstance(attributes, list):
            attributes = [attributes]

        mod = import_module(module_name)
        for attr in attributes:
            outer_globals[attr] = getattr(mod, attr)

    def global_import(module_name, short=None):
        if short is None:
            short = module_name
        outer_globals[module_name] = import_module(module_name)

    def global_import_star(module_name):
        mod = import_module(module_name)

        if '__all__' in mod.__dict__:
            names = mod.__dict__['__all__']
        else:
            names = [x for x in mod.__dict__ if not x.startswith('_') and x not in globals()]

        outer_globals.update({k: getattr(mod, k) for k in names})

    global_import('xarray', 'xr')
    global_import('scipy')
    global_import('numpy', 'np')
    global_import('pandas', 'pd')

    global_import('matplotlib.pyplot', 'plt')

    global_import_from('matplotlib', 'colors')

    global_import('scipy.ndimage', 'ndi')

    import os.path
    outer_globals['os'] = os

    global_import('bokeh')
    global_import_from('arpes.config', ['CONFIG', 'FIGURE_PATH'])

    global_import_from('arpes.laue', 'load_laue')

    global_import_star('arpes.analysis')
    global_import_star('arpes.fits')
    global_import_star('arpes.io')

    global_import_star('arpes.plotting')

    try:
        global_import_from('arpes.plotting.qt_tool', 'qt_tool')
    except ImportError:
        print('You will need to install PyQt5 in order to make `qt_tool` available.')

    global_import_star('arpes.plotting.annotations')

    global_import_star('arpes.repair')
    global_import_star('arpes.preparation')
    global_import_star('arpes.bootstrap')
    global_import_from('arpes.utilities.autoprep', 'prepare_raw_files')

    global_import_star('arpes.analysis.band_analysis_utils')
    global_import_from('arpes.corrections', 'fermi_edge_corrections')

    import arpes.xarray_extensions  # Ensure that extensions get loaded

    global_import_from('arpes.endstations', 'load_scan')
    global_import_from('arpes.utilities', ['clean_xlsx_dataset', 'default_dataset', 'swap_reference_map'])
    global_import_star('arpes.utilities.conversion')
    global_import_star('arpes.utilities.conversion.forward')
    global_import_from('arpes.endstations', ['endstation_from_alias', 'endstation_name_from_alias'])

    try:
        global_import_from('bokeh.io', 'output_notebook')
        from bokeh.io import output_notebook
    except ImportError:
        pass

    global_import('matplotlib')
    global_import_from('mpl_toolkits.axes_grid1.inset_locator', 'inset_axes')

    try:
        import matplotlib
        matplotlib.rcParams['animation.html'] = 'html5'
    except Exception: # pylint: disable=broad-except
        pass

    arpes.config.attempt_determine_workspace(permissive=True)
    arpes.config.load_plugins()

    outer_globals['ld'] = outer_globals['simple_load']

    global_import_from('pathlib', 'Path')
    global_import_from('arpes.config', 'use_tex')
    global_import('importlib')



import numpy as np

from arpes.fits import LinearModel
from arpes.typing import DataType

__all__ = ('kfermi_from_mdcs',)


def kfermi_from_mdcs(mdc_results: DataType, param=None):
    """
    Calculates a Fermi momentum using a series of MDCs and the known
    Fermi level (eV=0). This is especially useful to
    isolate an area for analysis.

    This method tolerates data that came from a prefixed model fit,
    but will always attempt to look for an attribute containing "center".

    :param mdc_results: A DataArray or Dataset containing `lmfit.ModelResult`s.
    :param param:
    :return:
    """
    real_param_name = None

    param_names = mdc_results.results.F.parameter_names

    if param in param_names and param is not None:
        real_param_name = param
    else:
        best_names = [p for p in param_names if 'center' in p]
        if param is not None:
            best_names = [p for p in best_names if param in p]

        assert len(best_names) == 1
        real_param_name = best_names[0]

    def nan_sieve(_, x):
        return not np.isnan(x.item())

    return LinearModel().guess_fit(
        mdc_results.F.p(real_param_name).T.filter_coord('eV', nan_sieve)).eval(x=0)



import collections
import copy
import functools

import numpy as np

import xarray as xr
from arpes.provenance import update_provenance

__all__ = ('replace_coords', 'disambiguate_coordinates',)


def disambiguate_coordinates(datasets, possibly_clashing_coordinates):
    coords_set = collections.defaultdict(list)
    for d in datasets:
        for c in possibly_clashing_coordinates:
            if c in d.coords:
                coords_set[c].append(d.coords[c])

    conflicted = []
    for c in possibly_clashing_coordinates:
        different_coords = coords_set[c]
        if not different_coords:
            continue

        if not functools.reduce(lambda x, y: (
                    np.array_equal(x[1], y) and x[0], y),
                                different_coords, (True, different_coords[0]))[0]:
            conflicted.append(c)

    after_deconflict = []
    for d in datasets:
        spectrum_name = list(d.data_vars.keys())[0]
        to_rename = {name: name + '-' + spectrum_name for name in d.dims if name in conflicted}
        after_deconflict.append(d.rename(to_rename))

    return after_deconflict


@update_provenance('Replace coordinates')
def replace_coords(arr: xr.DataArray, new_coords, mapping):
    coords = dict(copy.deepcopy(arr.coords))
    dims = list(copy.deepcopy(arr.dims))
    for old_dim, new_dim in mapping:
        coords[new_dim] = new_coords[new_dim]
        del coords[old_dim]
        dims[dims.index(old_dim)] = new_dim

    return xr.DataArray(
        arr.values,
        coords,
        dims,
        attrs=arr.attrs,
    )



from arpes.typing import DataType

__all__ = ('negate_energy',)

def negate_energy(data: DataType):
    if 'eV' in data.coords:
        data.coords['eV'].values = -data.coords['eV'].values
    return data



import numpy as np

import xarray as xr
from arpes.provenance import update_provenance

__all__ = ('infer_center_pixel', 'stitch_maps', )

def infer_center_pixel(arr: xr.DataArray):
    import arpes.xarray_extensions # pylint: disable=unused-import

    edges = arr.S.find_spectrum_angular_edges()
    return float((np.max(edges) + np.min(edges)) / 2 + np.min(arr.coords['pixel']))


@update_provenance('Stitch maps together')
def stitch_maps(arr: xr.DataArray, arr2: xr.DataArray, dimension='beta'):
    """
    Stitches together two maps by appending and potentially dropping frames in the first dataset.

    This is useful for beamline work when the beam is lost or in L-ARPES if laser output is blocked for part
    of a scan and a subsequent scan was taken to repair the problem.
    :param arr: Incomplete map
    :param arr2: completion of first map
    :return:
    """

    # as a first step we need to align the coords of map2 onto those of map1
    coord2 = arr2.coords[dimension].data.copy()
    coord1 = arr.coords[dimension].data.copy()

    first_repair_coordinate = coord2.data[0]
    i, lower, higher = None, None, None

    # search for the breakpoint
    for i, (lower, higher,) in enumerate(zip(coord1, coord1[1:])):
        if higher > first_repair_coordinate:
            break

    delta_low, delta_high = lower - first_repair_coordinate, higher - first_repair_coordinate
    if abs(delta_low) < abs(delta_high):
        delta = delta_low
    else:
        delta = delta_high
        i += 1

    shifted_repair_map = arr2.copy()
    shifted_repair_map.coords[dimension].data += delta

    attrs = arr.attrs.copy()
    good_data_slice = dict()
    good_data_slice[dimension] = slice(None, i)

    selected = arr.isel(**good_data_slice)
    selected.attrs.clear()
    shifted_repair_map.attrs.clear()
    concatted = xr.concat([selected, shifted_repair_map], dim=dimension)
    return xr.DataArray(
        concatted.data,
        concatted.coords,
        concatted.dims,
        attrs=attrs,
    )



import copy
import os
from pathlib import Path

import h5py
import numpy as np

import xarray as xr
from arpes.endstations import SESEndstation
from arpes.load_pxt import read_single_pxt
from arpes.provenance import provenance_from_file
from arpes.repair import negate_energy

__all__ = ('IgorExportEndstation',)


class IgorExportEndstation(SESEndstation):  # , FITSEndstation):
    """
    The Igor Export "Endstation" for ARPES data
    """

    PRINCIPAL_NAME = 'Igor'
    ALIASES = ['igor', 'igor-export', ]

    RENAME_KEYS = {}

    def load_single_frame(self, frame_path: str = None, scan_desc: dict = None, **kwargs):
        name, ext = os.path.splitext(frame_path)

        if 'nc' in ext or 'h5' in ext:
            # was converted to hdf5/NetCDF format with Conrad's Igor scripts
            scan_desc = copy.deepcopy(scan_desc)
            scan_desc['path'] = frame_path
            return self.load_SES_h5(scan_desc=scan_desc, robust_dimension_labels=True, **kwargs)

        # it's given by SES PXT files
        pxt_data = negate_energy(read_single_pxt(frame_path))
        return xr.Dataset({'spectrum': pxt_data}, attrs=pxt_data.attrs)

    def load_SES_h5(self, scan_desc: dict = None, robust_dimension_labels=False, **kwargs):
        """
        Imports an hdf5 dataset exported from Igor that was originally generated by a Scienta spectrometer
        in the SESb format. In order to understand the structure of these files have a look at Conrad's
        saveSESDataset in Igor Pro.

        :param scan_desc: Dictionary with extra information to attach to the xr.Dataset, must contain the location
        of the file
        :return: xr.Dataset
        """

        scan_desc = copy.deepcopy(scan_desc)

        data_loc = scan_desc.get('path', scan_desc.get('file'))
        p = Path(data_loc)
        if not p.exists():
            import arpes.config
            data_loc = os.path.join(arpes.config.DATA_PATH, data_loc)

        # wave_note = shim_wave_note(data_loc)
        wave_note = ""
        f = h5py.File(data_loc, 'r')

        primary_dataset_name = list(f)[0]
        # This is bugged for the moment in h5py due to an inability to read fixed length unicode strings
        # wave_note = f['/' + primary_dataset_name].attrs['IGORWaveNote']

        dimension_labels = list(f['/' + primary_dataset_name].attrs['IGORWaveDimensionLabels'][0])
        # print(list(f['/' + primary_dataset_name].attrs.keys()))

        if any(x == '' for x in dimension_labels):
            # print(dimension_labels)

            if not robust_dimension_labels:
                raise ValueError('Missing dimension labels. Use robust_dimension_labels=True to override')
            else:
                used_blanks = 0
                for i in range(len(dimension_labels)):
                    if dimension_labels[i] == '':
                        dimension_labels[i] = 'missing{}'.format(used_blanks)
                        used_blanks += 1

                        # print(dimension_labels)

        scaling = f['/' + primary_dataset_name].attrs['IGORWaveScaling'][-len(dimension_labels):]
        raw_data = f['/' + primary_dataset_name][:]

        scaling = [np.linspace(scale[1], scale[1] + scale[0] * raw_data.shape[i], raw_data.shape[i])
                   for i, scale in enumerate(scaling)]

        dataset_contents = {}
        attrs = scan_desc.pop('note', {})
        attrs.update(wave_note)

        built_coords = dict(zip(dimension_labels, scaling))

        deg_to_rad_coords = {'theta', 'beta', 'phi'}

        # the hemisphere axis is handled below
        built_coords = {k: c * (np.pi / 180) if k in deg_to_rad_coords else c
                        for k, c in built_coords.items()}

        deg_to_rad_attrs = {'theta', 'beta', 'alpha', 'chi'}
        for angle_attr in deg_to_rad_attrs:
            if angle_attr in attrs:
                attrs[angle_attr] = float(attrs[angle_attr]) * np.pi / 180

        dataset_contents['spectrum'] = xr.DataArray(
            raw_data,
            coords=built_coords,
            dims=dimension_labels,
            attrs=attrs,
        )

        provenance_from_file(dataset_contents['spectrum'], data_loc, {
            'what': 'Loaded SES dataset from HDF5.',
            'by': 'load_SES'
        })

        return xr.Dataset(
            dataset_contents,
            attrs={**scan_desc, 'name': primary_dataset_name},
        )



from collections import OrderedDict
import pandas as pd

from arpes.analysis import apply_mask
from arpes.corrections.background import remove_incoherent_background
from arpes.corrections.fermi_edge_corrections import apply_quadratic_fermi_edge_correction
from arpes.typing import DataType
from arpes.utilities import normalize_to_dataset, normalize_to_spectrum, deep_equals
from .fermi_edge_corrections import *
from .cycle import *


from arpes.io import simple_load

__all__ = ('build_reference_set', 'reference_key', 'apply_from_reference_set',
           'correction_from_reference_set',)


class HashableDict(OrderedDict):
    """
    Implements hashing for ordered dictionaries. The dictionary
    must be ordered for the hash to be stable.
    """
    def __hash__(self):
        return hash(frozenset(self.items()))


def reference_key(data: DataType):
    data = normalize_to_dataset(data)

    return HashableDict(data.S.reference_settings)


def correction_from_reference_set(data: DataType, reference_set):
    data = normalize_to_dataset(data)

    correction = None
    for k, corr in reference_set.items():
        if deep_equals(dict(reference_key(data)), dict(k)):
            correction = corr
            break

    return correction


def apply_from_reference_set(data: DataType, reference_set, **kwargs):
    correction = correction_from_reference_set(data, reference_set)

    data = normalize_to_spectrum(data)
    if correction is None:
        return data

    return apply_quadratic_fermi_edge_correction(data, correction, **kwargs)


def build_reference_set(df: pd.DataFrame, mask=None):
    references = {}

    for index, col in df.iterrows():
        print(index)
        data = simple_load(int(index), df)
        settings = data.S.reference_settings

        data = data.spectrum.S.sum_other(['phi', 'eV'])

        if mask is not None:
            data = apply_mask(data, mask, radius=-15)

        data = remove_incoherent_background(data)
        correction = build_quadratic_fermi_edge_correction(arr=data, fit_limit=0.0001)

        references[HashableDict(settings)] = correction

    return references



# pylint: disable=no-member

import copy
import itertools
import os.path
import warnings

import h5py
import numpy as np
from astropy.io import fits

import arpes.config
import xarray as xr
from arpes.endstations import EndstationBase, find_clean_coords
from arpes.provenance import provenance_from_file
from arpes.utilities import rename_keys

__all__ = ('SpinToFEndstation',)


class SpinToFEndstation(EndstationBase):
    PRINCIPAL_NAME = 'ALG-SToF'
    ALIASES = ['ALG-SToF', 'SToF', 'Spin-ToF', 'ALG-SpinToF']
    SKIP_ATTR_FRAGMENTS = {
        'MMX', 'TRVAL', 'TRDELT', 'COMMENT', 'OFFSET', 'SMOTOR', 'TUNIT', 'PMOTOR',
        'TDESC', 'NAXIS', 'TTYPE', 'TFORM', 'XTENSION', 'BITPIX', 'TDELT',
        'TRPIX',
    }

    COLUMN_RENAMINGS = {
        'TempA': 'temperature_cryo',
        'TempB': 'temperature_sample',
        'Current': 'photocurrent',
        'ALS_Beam_mA': 'beam_current',
        'Energy_Spectra': 'spectrum',
        'targetPlus': 't_up',
        'targetMinus': 't_down',
        'wave': 'spectrum',  # this should not occur simultaneously with 'Energy_Spectra'
        'Time_Target_Up': 't_up',
        'Time_Target_Down': 't_down',
        'Energy_Target_Up': 'up',
        'Energy_Target_Down': 'down',
        'Photocurrent_Up': 'photocurrent_up',
        'Photocurrent_Down': 'photocurrent_down',
        'Phi': 'phi',
    }

    RENAME_KEYS = {
        'LMOTOR0': 'x',
        'LMOTOR1': 'y',
        'LMOTOR2': 'z',
        'LMOTOR3': 'theta',
        'LMOTOR4': 'beta',
        'LMOTOR5': 'chi',
        'LMOTOR6': 'delay',
        'Phi': 'phi',
    }

    def load_SToF_hdf5(self, scan_desc: dict=None, **kwargs):
        """
        Imports a FITS file that contains ToF spectra.

        :param scan_desc: Dictionary with extra information to attach to the xr.Dataset, must contain the location
        of the file
        :return: xr.Dataset
        """

        scan_desc = copy.deepcopy(scan_desc)

        data_loc = scan_desc.get('path', scan_desc.get('file'))
        data_loc = data_loc if data_loc.startswith('/') else os.path.join(arpes.config.DATA_PATH, data_loc)

        f = h5py.File(data_loc, 'r')

        dataset_contents = dict()
        raw_data = f['/PRIMARY/DATA'][:]
        raw_data = raw_data[:, ::-1]  # Reverse the timing axis
        dataset_contents['raw'] = xr.DataArray(
            raw_data,
            coords={'x_pixels': np.linspace(0, 511, 512),
                    't_pixels': np.linspace(0, 511, 512)},
            dims=('x_pixels', 't_pixels'),
            attrs=f['/PRIMARY'].attrs.items(),
        )

        provenance_from_file(dataset_contents['raw'], data_loc, {
            'what': 'Loaded Anton and Ping DLD dataset from HDF5.',
            'by': 'load_DLD',
        })

        return xr.Dataset(
            dataset_contents,
            attrs=scan_desc
        )

    def load_SToF_fits(self, scan_desc: dict=None, **kwargs):
        scan_desc = dict(copy.deepcopy(scan_desc))

        data_loc = scan_desc.get('path', scan_desc.get('file'))
        data_loc = data_loc if data_loc.startswith('/') else os.path.join(arpes.config.DATA_PATH, data_loc)

        hdulist = fits.open(data_loc)

        hdulist[0].verify('fix+warn')
        header_hdu, hdu = hdulist[0], hdulist[1]

        scan_desc.update(dict(hdu.header))
        scan_desc.update(dict(header_hdu.header))

        drop_attrs = ['COMMENT', 'HISTORY', 'EXTEND', 'SIMPLE', 'SCANPAR', 'SFKE_0']
        for dropped_attr in drop_attrs:
            if dropped_attr in scan_desc:
                del scan_desc[dropped_attr]

        coords, dimensions, spectrum_shape = find_clean_coords(hdu, scan_desc)
        dimensions = {k: [SpinToFEndstation.RENAME_KEYS.get(n, n) for n in v] for k, v in dimensions.items()}
        coords = rename_keys(coords, SpinToFEndstation.RENAME_KEYS)

        columns = hdu.columns

        spin_column_names = {'targetMinus', 'targetPlus', 'Time_Target_Up', 'Time_Target_Down', 'Energy_Target_Up',
                             'Energy_Target_Down', 'Photocurrent_Up', 'Photocurrent_Down'}

        is_spin_resolved = any(cname in columns.names for cname in spin_column_names)
        spin_columns = ['Current' 'TempA', 'TempB', 'ALS_Beam_mA'] + list(spin_column_names)
        straight_columns = ['Current', 'TempA', 'TempB', 'ALS_Beam_mA', 'Energy_Spectra', 'wave']
        take_columns = spin_columns if is_spin_resolved else straight_columns

        # We could do our own spectrum conversion too, but that would be more annoying
        # it would slightly improve accuracy though
        spectra_names = [name for name in take_columns if name in columns.names]

        skip_predicates = {lambda k: any(s in k for s in self.SKIP_ATTR_FRAGMENTS), }

        scan_desc = {k: v for k, v in scan_desc.items()
                    if not any(pred(k) for pred in skip_predicates)}
        scan_desc = rename_keys(scan_desc, SpinToFEndstation.RENAME_KEYS)

        # TODO, we should try to unify this with the FITS file loader, but there are a few current inconsistencies
        data_vars = {}

        for spectrum_name in spectra_names:
            column_shape = spectrum_shape[spectrum_name]
            data_for_resize = hdu.data.columns[spectrum_name].array

            try:
                # best possible case is that we have identically all of the data
                resized_data = data_for_resize.reshape(column_shape)
            except ValueError:
                # if we stop scans early, the header is already written and so the size of the data will be small along
                # the experimental axes
                rest_column_shape = column_shape[1:]
                n_per_slice = int(np.prod(rest_column_shape))
                total_shape = data_for_resize.shape
                total_n = np.prod(total_shape)

                n_slices = total_n // n_per_slice

                if (total_n // n_per_slice != total_n / n_per_slice):
                    # the last slice was in the middle of writing when something hit the fan
                    # we need to infer how much of the data to read, and then repeat the above
                    # we need to cut the data

                    # This can happen when the labview crashes during data collection,
                    # we use column_shape[1] because of the row order that is used in the FITS file
                    data_for_resize = data_for_resize[0:(total_n // n_per_slice) * column_shape[1]]
                    warnings.warn(
                        'Column {} was in the middle of slice when DAQ stopped. Throwing out incomplete slice...'.format(
                            spectrum_name))

                column_shape = list(column_shape)
                column_shape[0] = n_slices

                try:
                    resized_data = data_for_resize.reshape(column_shape)
                except Exception:
                    # we should probably zero pad in the case where the slices are not the right size
                    continue

                altered_dimension = dimensions[spectrum_name][0]
                coords[altered_dimension] = coords[altered_dimension][:n_slices]

            data_vars[spectrum_name] = (dimensions[spectrum_name], resized_data, scan_desc,)

        data_vars = rename_keys(data_vars, SpinToFEndstation.COLUMN_RENAMINGS)
        if 'beam_current' in data_vars and np.all(data_vars['beam_current'][1] == 0):
            # Wasn't taken at a beamline
            del data_vars['beam_current']

        hdulist.close()

        relevant_dimensions = {k for k in coords.keys() if k in
                               set(itertools.chain(*[l[0] for l in data_vars.values()]))}
        relevant_coords = {k: v for k, v in coords.items() if k in relevant_dimensions}

        dataset = xr.Dataset(
            data_vars,
            relevant_coords,
            scan_desc,
        )

        for var_name, data_arr in dataset.data_vars.items():
            if 'time' in data_arr.dims:
                data_arr.data = data_arr.sel(time=slice(None, None, -1)).data

        provenance_from_file(dataset, data_loc, {
            'what': 'Loaded Spin-ToF dataset',
            'by': 'load_DLD',
        })

        return dataset

    def load(self, scan_desc: dict=None, **kwargs):
        if scan_desc is None:
            warnings.warn('Attempting to make due without user associated scan_desc for the file')
            raise TypeError('Expected a dictionary of scan_desc with the location of the file')

        data_loc = scan_desc.get('path', scan_desc.get('file'))
        scan_desc = {k: v for k, v in scan_desc.items() if not isinstance(v, float) or not np.isnan(v)}

        if os.path.splitext(data_loc)[1] == '.fits':
            return self.load_SToF_fits(scan_desc)

        return self.load_SToF_hdf5(scan_desc)



import copy
import functools

import numpy as np
from scipy.ndimage import geometric_transform

import xarray as xr
from arpes.provenance import provenance, update_provenance
from arpes.typing import DataType
from arpes.utilities import lift_dataarray_to_generic
from arpes.utilities.normalize import normalize_to_spectrum

__all__ = ('flip_axis', 'normalize_dim', 'dim_normalizer', 'transform_dataarray_axis',
           'normalize_total', 'sort_axis',)


@update_provenance('Sort Axis')
def sort_axis(data: xr.DataArray, axis_name):
    assert isinstance(data, xr.DataArray)
    copied = data.copy(deep=True)
    coord = data.coords[axis_name].values
    order = np.argsort(coord)
    print(order)
    copied.values = np.take(copied.values, order, axis=list(data.dims).index(axis_name))
    copied.coords[axis_name] = np.sort(copied.coords[axis_name])
    return copied


@update_provenance('Flip data along axis')
def flip_axis(arr: xr.DataArray, axis_name, flip_data=True):
    coords = copy.deepcopy(arr.coords)
    coords[axis_name] = coords[axis_name][::-1]

    return xr.DataArray(
        np.flip(arr.values, arr.dims.index(axis_name)) if flip_data else arr.values,
        coords,
        arr.dims,
        attrs=arr.attrs
    )


def soft_normalize_dim(arr: xr.DataArray, dim_or_dims, keep_id=False, amp_limit=100):
    dims = dim_or_dims
    if isinstance(dim_or_dims, str):
        dims = [dims]

    summed_arr = arr.fillna(arr.mean()).sum([d for d in arr.dims if d not in dims])
    normalized_arr = arr / (summed_arr / np.product(summed_arr.shape))

    to_return = xr.DataArray(
        normalized_arr.values,
        arr.coords,
        arr.dims,
        attrs=arr.attrs
    )

    if not keep_id and 'id' in to_return.attrs:
        del to_return.attrs['id']

    provenance(to_return, arr, {
        'what': 'Normalize axis or axes',
        'by': 'normalize_dim',
        'dims': dims,
    })

    return to_return


@lift_dataarray_to_generic
def normalize_dim(arr: DataType, dim_or_dims, keep_id=False):
    """
    Normalizes the intensity so that all values along arr.sum(dims other than those in ``dim_or_dims``)
    have the same value. The function normalizes so that the average value of cells in
    the output is 1.
    :param dim_name:
    :return:
    """

    dims = dim_or_dims
    if isinstance(dim_or_dims, str):
        dims = [dims]

    summed_arr = arr.fillna(arr.mean()).sum([d for d in arr.dims if d not in dims])
    normalized_arr = arr / (summed_arr / np.product(summed_arr.shape))

    to_return = xr.DataArray(
        normalized_arr.values,
        arr.coords,
        arr.dims,
        attrs=arr.attrs
    )

    if not keep_id and 'id' in to_return.attrs:
        del to_return.attrs['id']

    provenance(to_return, arr, {
        'what': 'Normalize axis or axes',
        'by': 'normalize_dim',
        'dims': dims,
    })

    return to_return


@update_provenance('Normalize total spectrum intensity')
def normalize_total(data: DataType):
    data = normalize_to_spectrum(data)

    return data / (data.sum(data.dims) / 1000000)


def dim_normalizer(dim_name):
    def normalize(arr: xr.DataArray):
        if dim_name not in arr.dims:
            return arr
        return normalize_dim(arr, dim_name)

    return normalize


def transform_dataarray_axis(f, old_axis_name: str,
                             new_axis_name: str, new_axis,
                             dataset: xr.DataArray, prep_name,
                             transform_spectra=None, remove_old=True):

    ds = dataset.copy()
    if transform_spectra is None:
        # transform *all* DataArrays in the dataset that have old_axis_name in their dimensions
        transform_spectra = {k: v for k, v in ds.data_vars.items()
                             if old_axis_name in v.dims}

    ds.coords[new_axis_name] = new_axis

    new_dataarrays = []
    for name in transform_spectra.keys():
        dr = ds[name]

        old_axis = dr.dims.index(old_axis_name)
        shape = list(dr.sizes.values())
        shape[old_axis] = len(new_axis)
        new_dims = list(dr.dims)
        new_dims[old_axis] = new_axis_name

        g = functools.partial(f, axis=old_axis)
        output = geometric_transform(dr.values, g, output_shape=shape, output='f', order=1)

        new_coords = dict(dr.coords)
        new_coords.pop(old_axis_name)

        new_dataarray = xr.DataArray(output, coords=new_coords, dims=new_dims,
                                     attrs=dr.attrs.copy(), name=prep_name(dr.name))
        new_dataarrays.append(new_dataarray)
        if 'id' in new_dataarray.attrs:
            del new_dataarray.attrs['id']

        if remove_old:
            del ds[name]
        else:
            assert(prep_name(name) != name and "You must make sure names don't collide")

    new_ds = xr.merge([
        ds,
        *new_dataarrays
    ])

    new_ds.attrs.update(ds.attrs.copy())

    if 'id' in new_ds:
        del new_ds.attrs['id']

    provenance(new_ds, dataset, {
        'what': 'Transformed a Dataset coordinate axis',
        'by': 'transform_dataarray_axis',
        'old_axis': old_axis_name,
        'new_axis': new_axis_name,
        'transformed_vars': list(transform_spectra.keys()),
    })

    return new_ds



from .annotations import *

from .bands import *
from .basic import *
from .dispersion import *
from .fermi_edge import *
from .fermi_surface import *
from .dos import *
from .parameter import *

from .stack_plot import *

from .spin import *
from .tarpes import *
from .visualize_3d import *
from .spatial import *

from .movie import *

# 'Tools'
# Note, we lift Bokeh imports into definitions in case people don't want to install Bokeh
# and also because of an undesirable interaction between pytest and Bokeh due to Bokeh's use
# of jinja2.
from .interactive import *
from .band_tool import *
from .comparison_tool import *
from .curvature_tool import *
from .fit_inspection_tool import *
from .mask_tool import *
from .path_tool import *
from .dyn_tool import *

from .utils import savefig
from .utils import remove_colorbars



import warnings
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np

import xarray as xr
from arpes.io import load_dataset
from arpes.preparation import normalize_dim
from arpes.provenance import save_plot_provenance
from arpes.utilities import bz
from arpes.utilities.conversion import remap_coords_to

from .utils import path_for_plot, label_for_dim, label_for_symmetry_point, label_for_colorbar

__all__ = ['plot_dispersion', 'labeled_fermi_surface',
           'cut_dispersion_plot', 'fancy_dispersion', 'reference_scan_fermi_surface',
           'hv_reference_scan', 'scan_var_reference_plot']


def band_path(band):
    import holoviews as hv  # pylint: disable=import-error
    return hv.Path([band.center.values, band.coords[band.dims[0]].values])


@save_plot_provenance
def plot_dispersion(spectrum: xr.DataArray, bands, out=None):
    ax = spectrum.plot()

    for band in bands:
        plt.scatter(band.center.values, band.coords[band.dims[0]].values)

    if out is not None:
        filename = path_for_plot(out)
        plt.savefig(filename)
        return filename
    else:
        return ax


@save_plot_provenance
def cut_dispersion_plot(data: xr.DataArray, e_floor=None, title=None, ax=None, include_symmetry_points=True,
                        out=None, quality='high', **kwargs):
    """
    Makes a 3D cut dispersion plot. At the moment this only supports rectangular BZs.
    :param data:
    :param e_floor:
    :param title:
    :param ax:
    :param out:
    :param kwargs:
    :return:
    """

    # to get nice labeled edges you could use shapely
    sampling = {
        'paper': 400,
        'high': 100,
        'low': 40,
    }.get(quality, 100)

    assert 'eV' in data.dims
    assert e_floor is not None

    new_dim_order = list(data.dims)
    new_dim_order.remove('eV')
    new_dim_order = new_dim_order + ['eV']
    data = data.transpose(*new_dim_order)

    # prep data to be used for rest of cuts
    lower_part = data.sel(eV=slice(None, 0))
    floor = lower_part.S.fat_sel(eV=e_floor)

    bz_mask = bz.reduced_bz_mask(lower_part, scale_zone=True)
    left_mask = bz.reduced_bz_E_mask(lower_part, 'X', e_floor, scale_zone=True)
    right_mask = bz.reduced_bz_E_mask(lower_part, 'Y', e_floor, scale_zone=True)
    mask_for = lambda x: left_mask if x.shape == left_mask.shape else right_mask

    x_dim, y_dim, z_dim = tuple(new_dim_order)
    x_coords, y_coords, _ = data.coords[x_dim], data.coords[y_dim], data.coords[z_dim] # x_coords, y_coords, z_coords

    if ax is None:
        fig = plt.figure(figsize=(7, 7))
        ax = fig.add_subplot(1, 1, 1, projection='3d')

    if title is None:
        title = '{} Cut Through Symmetry Points'.format(data.S.label.replace('_', ' '))

    ax.set_title(title)
    colormap = plt.get_cmap('Blues')

    # color fermi surface
    fermi_surface = data.S.fermi_surface
    Xs, Ys = np.meshgrid(x_coords, y_coords)

    Zs = np.zeros(fermi_surface.data.shape)
    Zs[bz_mask] = np.nan
    scale_colors = max(np.max(fermi_surface.data), np.max(floor.data))
    ax.plot_surface(Xs, Ys, Zs.T, facecolors=colormap(fermi_surface.data.T / scale_colors), shade=False,
                    vmin=-1, vmax=1, antialiased=True, rcount=sampling, ccount=sampling)

    # color right edge
    right_sel = dict()
    edge_val = np.max(lower_part.coords[x_dim].data)
    right_sel[x_dim] = edge_val
    right_edge = lower_part.S.fat_sel(**right_sel)

    Ys, Zs = np.meshgrid(lower_part.coords[y_dim], lower_part.coords[z_dim])
    Xs = np.ones(right_edge.shape) * edge_val
    Xs[mask_for(Xs)] = np.nan
    ax.plot_surface(Xs.T, Ys, Zs, facecolors=colormap(right_edge.data.T / scale_colors),
                    vmin=-1, vmax=1, shade=False, antialiased=True, rcount=sampling, ccount=sampling)

    # color left edge
    left_sel = dict()
    edge_val = np.min(lower_part.coords[y_dim].data)
    left_sel[y_dim] = edge_val
    left_edge = lower_part.S.fat_sel(**left_sel)
    # max_left_edge = np.max(left_edge.data)

    Xs, Zs = np.meshgrid(lower_part.coords[x_dim], lower_part.coords[z_dim])
    Ys = np.ones(left_edge.shape) * edge_val
    Ys[mask_for(Ys)] = np.nan
    ax.plot_surface(Xs, Ys.T, Zs, facecolors=colormap(left_edge.data.T / scale_colors),
                    vmin=-1, vmax=1, antialiased=True, shade=False, rcount=sampling, ccount=sampling)

    # selection region
    # floor
    Xs, Ys = np.meshgrid(floor.coords[x_dim], floor.coords[y_dim])
    Zs = np.ones(floor.data.shape) * e_floor
    Zs[np.logical_not(bz_mask)] = np.nan
    ax.plot_surface(Xs, Ys, Zs.T, facecolors=colormap(floor.data.T / scale_colors),
                    vmin=-1, vmax=1, shade=False, antialiased=True, rcount=sampling, ccount=sampling)

    # determine the axis along X, Y
    axis_X = bz.axis_along(data, 'X')
    axis_Y = bz.axis_along(data, 'Y')

    # left and right inset faces
    inset_face = lower_part.S.along(['G', 'X'], axis_name=axis_X, extend_to_edge=True).sel(eV=slice(e_floor, None))
    Xs, Zs = np.meshgrid(inset_face.coords[axis_X], inset_face.coords[z_dim])
    Ys = np.ones(inset_face.data.shape)
    if y_dim == axis_X:
        Ys *= data.S.phi_offset
        Xs += data.S.map_angle_offset
        Xs, Ys = Ys, Xs
    else:
        Ys *= data.S.map_angle_offset
        Xs += data.S.phi_offset
    ax.plot_surface(Xs, Ys, Zs, facecolors=colormap(inset_face.data / scale_colors),
                    shade=False, antialiased=True, zorder=1, rcount=sampling, ccount=sampling)

    inset_face = lower_part.S.along(['G', 'Y'], axis_name=axis_Y, extend_to_edge=True).sel(eV=slice(e_floor, None))
    Ys, Zs = np.meshgrid(inset_face.coords[axis_Y], inset_face.coords[z_dim])
    Xs = np.ones(inset_face.data.shape)
    if x_dim == axis_Y:
        Xs *= data.S.map_angle_offset
        Ys += data.S.phi_offset
        Xs, Ys = Ys, Xs
    else:
        Xs *= data.S.phi_offset
        Ys += data.S.map_angle_offset

    ax.plot_surface(Xs, Ys, Zs, facecolors=colormap(inset_face.data / scale_colors),
                    shade=False, antialiased=True, zorder=1, rcount=sampling, ccount=sampling)

    ax.set_xlabel(label_for_dim(data, x_dim))
    ax.set_ylabel(label_for_dim(data, y_dim))
    ax.set_zlabel(label_for_dim(data, z_dim))

    zlim = ax.get_zlim3d()
    if include_symmetry_points:
        for point_name, point_location in data.S.iter_symmetry_points:
            coords = [point_location.get(d, 0.02) for d in new_dim_order]
            ax.scatter(*zip(coords), marker='.', color='red', zorder=1000)
            coords[new_dim_order.index('eV')] += 0.1
            ax.text(*coords, label_for_symmetry_point(point_name), color='red', ha='center', va='top')

    ax.set_zlim3d(*zlim)
    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()


@save_plot_provenance
def hv_reference_scan(data, out=None, e_cut=-0.05, bkg_subtraction=0.8, **kwargs):
    fs = data.S.fat_sel(eV=e_cut)
    fs = normalize_dim(fs, 'hv', keep_id=True)
    fs.data -= bkg_subtraction * np.mean(fs.data)
    fs.data[fs.data < 0] = 0

    _, ax = labeled_fermi_surface(fs, hold=True, **kwargs)

    all_scans = data.attrs['df']
    all_scans = all_scans[all_scans.id != data.attrs['id']]
    all_scans = all_scans[(all_scans.spectrum_type != 'xps_spectrum') |
                          (all_scans.spectrum_type == 'hv_map')]

    scans_by_hv = defaultdict(list)
    for _, row in all_scans.iterrows():
        scan = load_dataset(row.id)

        scans_by_hv[round(scan.S.hv)].append(scan.S.label.replace('_', ' '))

    dim_order = ax.dim_order
    handles = []
    handle_labels = []

    prop_cycle = plt.rcParams['axes.prop_cycle']
    colors_cycle = prop_cycle.by_key()['color']

    for line_color, (hv, labels) in zip(colors_cycle, scans_by_hv.items()):
        full_label = '\n'.join(labels)

        # determine direction
        if dim_order[0] == 'hv':
            # photon energy is along the x axis, we want an axvline
            handles.append(ax.axvline(hv, label=full_label, color=line_color))
        else:
            # photon energy is along the y axis, we want an axhline
            handles.append(ax.axhline(hv, label=full_label, color=line_color))

        handle_labels.append(full_label)

    plt.legend(handles, handle_labels)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()

@save_plot_provenance
def reference_scan_fermi_surface(data, out=None, **kwargs):
    fs = data.S.fermi_surface
    _, ax = labeled_fermi_surface(fs, hold=True, **kwargs)

    referenced_scans = data.S.referenced_scans
    handles = []
    for index, row in referenced_scans.iterrows():
        scan = load_dataset(row.id)
        remapped_coords = remap_coords_to(scan, data)
        dim_order = ax.dim_order
        ls = ax.plot(remapped_coords[dim_order[0]], remapped_coords[dim_order[1]], label=index.replace('_', ' '))
        handles.append(ls[0])

    plt.legend(handles=handles)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()


@save_plot_provenance
def labeled_fermi_surface(data, title=None, ax=None, hold=False,
                          include_symmetry_points=True, include_bz=True,
                          out=None, fermi_energy=0, **kwargs):
    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 7))

    if title is None:
        title = '{} Fermi Surface'.format(data.S.label.replace('_', ' '))

    if 'eV' in data.dims:
        data = data.S.generic_fermi_surface(fermi_energy)

    mesh = data.plot(ax=ax)
    mesh.colorbar.set_label(label_for_colorbar(data))

    if data.S.is_differentiated:
        mesh.set_cmap('Blues')

    dim_order = [ax.get_xlabel(), ax.get_ylabel()]

    setattr(ax, 'dim_order', dim_order)
    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel(label_for_dim(data, ax.get_ylabel()))
    ax.set_title(title)

    marker_color = 'red' if data.S.is_differentiated else 'red'

    if include_bz:
        symmetry = bz.bz_symmetry(data.S.iter_own_symmetry_points)

        # TODO Implement this
        warnings.warn('BZ region display not implemented.')

    if include_symmetry_points:
        for point_name, point_location in data.S.iter_symmetry_points:
            warnings.warn('Symmetry point locations are not k-converted')
            coords = [point_location[d] for d in dim_order]
            ax.plot(*coords, marker='.', color=marker_color)
            ax.annotate(label_for_symmetry_point(point_name), coords, color=marker_color,
                        xycoords='data', textcoords='offset points', xytext=(0, -10),
                        va='top', ha='center', fontsize=14)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    if not hold:
        plt.show()
    else:
        return fig, ax


@save_plot_provenance
def fancy_dispersion(data, title=None, ax=None, out=None, include_symmetry_points=True,
                     norm=None, **kwargs):
    if ax is None:
        _, ax = plt.subplots(figsize=(8, 5))

    if title is None:
        title = data.S.label.replace('_', ' ')

    mesh = data.plot(norm=norm, ax=ax, **kwargs)
    mesh.colorbar.set_label(label_for_colorbar(data))

    if data.S.is_differentiated:
        mesh.set_cmap('Blues')

    original_x_label = ax.get_xlabel()
    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel(label_for_dim(data, ax.get_ylabel()))
    ax.set_title(title, fontsize=14)

    # This can probably be pulled out into a a helper
    marker_color = 'red' if data.S.is_differentiated else 'red'
    if include_symmetry_points:
        for point_name, point_locations in data.attrs.get('symmetry_points', {}).items():
            if not isinstance(point_locations, list):
                point_locations = [point_locations]
            for single_location in point_locations:
                coords = (single_location[original_x_label], ax.get_ylim()[1],)
                ax.plot(*coords, marker=11, color=marker_color)
                ax.annotate(label_for_symmetry_point(point_name), coords, color=marker_color,
                            xycoords='data', textcoords='offset points', xytext=(0, -10),
                            va='top', ha='center')


    ax.axhline(0, color='red', alpha=0.8, linestyle='--', linewidth=1)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()
    return ax


@save_plot_provenance
def scan_var_reference_plot(data, title=None, ax=None, norm=None, out=None, **kwargs):
    if ax is None:
        _, ax = plt.subplots(figsize=(8, 5))

    if title is None:
        title = data.S.label.replace('_', ' ')

    plot = data.plot(norm=norm, ax=ax)
    plot.colorbar.set_label(label_for_colorbar(data))

    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel(label_for_dim(data, ax.get_ylabel()))

    ax.set_title(title, font_size=14)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()



import warnings
from pathlib import Path

import numpy as np

import xarray as xr
from arpes.endstations import HemisphericalEndstation
from arpes.utilities import clean_keys

__all__ = ('MBSEndstation',)


class MBSEndstation(HemisphericalEndstation):
    """
    Implements loading text files from the MB Scientific text file format.

    There's not too much metadata here except what comes with the analyzer settings.
    """

    PRINCIPAL_NAME = 'MBS'
    ALIASES = ['MB Scientific',]

    RENAME_KEYS = {
        'deflx': 'psi',
    }

    def resolve_frame_locations(self, scan_desc: dict=None):
        return [scan_desc['path']]

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict=None):
        warnings.warn('Loading from text format misses metadata. You will need to supply '
                      'missing coordinates as appropriate.')
        data.attrs['psi'] = float(data.attrs['psi'])
        for s in data.S.spectra:
            s.attrs['psi'] = float(s.attrs['psi'])

        defaults = {'x': np.nan, 'y': np.nan, 'z': np.nan, 'theta': 0, 'beta': 0,
                    'chi': 0, 'alpha': np.nan, 'hv': np.nan}
        for k, v in defaults.items():
            data.attrs[k] = v
            for s in data.S.spectra:
                s.attrs[k] = v

        return super().postprocess_final(data, scan_desc)

    def load_single_frame(self, frame_path: str=None, scan_desc: dict=None, **kwargs):
        p = Path(frame_path)

        with open(str(p)) as f:
            lines = f.readlines()

        lines = [l.strip() for l in lines]
        data_index = lines.index('DATA:')
        header = lines[:data_index]
        data = lines[data_index + 1:]
        data = [d.split() for d in data]
        data = np.array([[float(f) for f in d] for d in data])

        header = [h.split('\t') for h in header]
        alt = [h for h in header if len(h) == 1]
        header = [h for h in header if len(h) == 2]
        header.append(['alt', str(alt)])
        attrs = clean_keys(dict(header))

        eV_axis = np.linspace(
            float(attrs['start_k_e_']),
            float(attrs['end_k_e_']),
            num=int(attrs['no_steps']),
            endpoint=False,
        )

        n_eV = int(attrs['no_steps'])
        idx_eV = data.shape.index(n_eV)

        if len(data.shape) == 2:
            phi_axis = np.linspace(
                float(attrs['xscalemin']),
                float(attrs['xscalemax']),
                num=data.shape[1 if idx_eV == 0 else 0],
                endpoint=False
            )

            coords = {'phi': phi_axis * np.pi / 180, 'eV': eV_axis}
            dims = ['eV', 'phi'] if idx_eV == 0 else ['phi', 'eV']
        else:
            coords = {'eV': eV_axis}
            dims = ['eV']


        return xr.Dataset({'spectrum': xr.DataArray(
            data, coords=coords, dims=dims, attrs=attrs
        )}, attrs=attrs)



import itertools

import matplotlib.pyplot as plt

from arpes.io import simple_load
from arpes.plotting import annotate_point
from arpes.plotting.utils import path_for_plot
from arpes.provenance import save_plot_provenance
from arpes.utilities import normalize_to_spectrum

__all__ = ('reference_scan_spatial',)


@save_plot_provenance
def reference_scan_spatial(data, out=None, **kwargs):
    data = normalize_to_spectrum(data)

    dims = [d for d in data.dims if d in {'cycle', 'phi', 'eV'}]

    summed_data = data.sum(dims, keep_attrs=True)

    fig, ax = plt.subplots(3, 2, figsize=(15, 15))
    flat_axes = list(itertools.chain(*ax))

    summed_data.plot(ax=flat_axes[0])
    flat_axes[0].set_title(r'Full \textbf{eV} range')

    dims_except_eV = [d for d in dims if d != 'eV']
    summed_data = data.sum(dims_except_eV)

    mul = 0.2
    rng = data.coords['eV'].max().item() - data.coords['eV'].min().item()
    offset = data.coords['eV'].max().item()
    if offset > 0:
        offset = 0

    if rng > 3:
        mul = rng / 5.

    for i in range(5):
        low_e, high_e = -mul * (i + 1) + offset, -mul * i + offset
        title = r'\textbf{eV}' + ': {:.2g} to {:.2g}'.format(low_e, high_e)
        summed_data.sel(eV=slice(low_e, high_e)).sum('eV').plot(ax=flat_axes[i+1])
        flat_axes[i + 1].set_title(title)

    y_range = flat_axes[0].get_ylim()
    x_range = flat_axes[0].get_xlim()
    delta_one_percent = ((x_range[1] - x_range[0]) / 100, (y_range[1] - y_range[0]) / 100)

    smart_delta = (2 * delta_one_percent[0], -1.5 * delta_one_percent[0])

    referenced = data.S.referenced_scans

    # idea here is to collect points by those that are close together, then
    # only plot one annotation
    condensed = []
    cutoff = 3 # 3 percent
    for index, row in referenced.iterrows():
        ff = simple_load(index)

        x, y, _ = ff.S.sample_pos
        found = False
        for cx, cy, cl in condensed:
            if abs(cx  - x) < cutoff * abs(delta_one_percent[0]) and abs(cy - y) < cutoff * abs(delta_one_percent[1]):
                cl.append(index)
                found = True
                break

        if not found:
            condensed.append((x, y, [index]))

    for fax in flat_axes:
        for cx, cy, cl in condensed:
            annotate_point(fax, (cx, cy,), ','.join([str(l) for l in cl]),
                           delta=smart_delta, fontsize='large')

    plt.tight_layout()

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax



import warnings

import pandas as pd

import xarray as xr
from arpes.io import simple_load
from arpes.pipelines import convert_scan_to_kspace
from arpes.preparation import normalize_dim
from arpes.utilities import default_dataset

__all__ = ['make_reference_plots']


def make_reference_plots(df: pd.DataFrame=None, with_kspace=False):
    if df is None:
        df = default_dataset()

    try:
        df = df[df.spectrum_type != 'xps_spectrum']
    except TypeError:
        warnings.warn('Unable to filter out XPS files, did you attach spectra type?')

    # Make scans indicating cut locations
    for index, row in df.iterrows():
        try:
            scan = simple_load(index)

            if isinstance(scan, xr.Dataset):
                # make plot series normalized by current:
                scan.S.reference_plot(out=True)
            else:
                scan.S.reference_plot(out=True, use_id=False)

                if scan.S.spectrum_type == 'spectrum':
                    # Also go and make a normalized version
                    normed = normalize_dim(scan, 'phi')
                    normed.S.reference_plot(out=True, use_id=False, pattern='{}_norm_phi.png')

                    if with_kspace:
                        kspace_converted = convert_scan_to_kspace(scan)
                        kspace_converted.S.reference_plot(out=True, use_id=False, pattern='k_{}.png')

                        normed_k = normalize_dim(kspace_converted, 'kp')
                        normed_k.S.reference_plot(out=True, use_id=False, pattern='k_{}_norm_kp.png')

        except Exception as e:
            print(str(e))
            warnings.warn('Cannot make plots for {}'.format(index))



"""
Provides general utility methods that get used during the course of analysis.
"""

import copy
import itertools
import json
import os
import re
import warnings
from math import sin, cos
from operator import itemgetter

import datetime

from arpes.typing import DataType

import pandas as pd
import numpy as np
import xarray as xr

from arpes import constants
from .funcutils import *
from .normalize import *
from .dataset import *
from .xarray import *
from .region import *
from .dict import *
from .attrs import *
from .collections import *


def enumerate_dataarray(arr: xr.DataArray):
    for coordinate in itertools.product(*[arr.coords[d] for d in arr.dims]):
        zip_location = dict(zip(arr.dims, (float(f) for f in coordinate)))
        yield zip_location, arr.loc[zip_location].values.item()


def fix_burnt_pixels(spectrum):
    """
    In reality the analyzers cannot provide perfect images for us. One of the
    principle failure modes is that individual pixels can get burnt out and will
    not provide any counts, or will provide consistently fewer or more than other
    pixels.

    Our approach here is to look for peaks in the difference across pixels and
    frames of a spectrum as indication of issues to be fixed. To patch the
    pixels, we replace them with the average value of their neighbors.

    spectrum - <npArray> containing the pixels

    returns: <npArray> containing the fixed pixels
    """
    pass


def jacobian_correction(energies, lattice_constant, theta, beta, alpha, phis, rhat):
    """
    Because converting from angles to momenta does not preserve area, we need
    to multiply by the Jacobian of the transformation in order to get the
    appropriate number of counts in the new cells.

    This differs across all the cells of a spectrum, because E and phi change.
    This function builds an array with the same shape that has the appropriate
    correction for each cell.

    energies - <npArray> the linear sampling of energies across the spectrum
    phis - <npArray> the linear sampling of angles across the spectrum

    returns: <npArray> a 2D array of the Jacobian correction to apply to each
    pixel in the spectrum
    """

    k0s = constants.K_INV_ANGSTROM * np.sqrt(energies) * lattice_constant / np.pi

    dkxdphi = (cos(theta) * cos(alpha) * np.cos(phis) -
               sin(theta) * np.sin(phis))

    dkydphi = (
        -cos(theta) * sin(beta) * np.sin(phis) +
        np.cos(phis) * (
            cos(beta) * sin(alpha) -
            cos(alpha) * sin(theta) * sin(beta)))

    # return the dot product
    rhat_x, rhat_y = rhat

    geometric_correction = rhat_x * dkxdphi + rhat_y * dkydphi
    return np.outer(k0s, geometric_correction)


def arrange_by_indices(items, indices):
    """
    This function is best illustrated by the example below. It arranges the
    items in the input according to the new indices that each item should occupy.

    It also has an inverse available in 'unarrange_by_indices'.

    Ex:
    arrange_by_indices(['a', 'b', 'c'], [1, 2, 0])
     => ['b', 'c', 'a']
    """
    return [items[i] for i in indices]


def unarrange_by_indices(items, indices):
    """
    The inverse function to 'arrange_by_indices'.

    Ex:
    unarrange_by_indices(['b', 'c', 'a'], [1, 2, 0])
     => ['a', 'b', 'c']
    """

    return [x for x, _ in sorted(zip(indices, items), key=itemgetter(0))]


ATTRS_MAP = {
    'PuPol': 'pump_pol',
    'PrPol': 'probe_pol',
    'SFLNM0': 'lens_mode',
    'Lens Mode': 'lens_mode',
    'Excitation Energy': 'hv',
    'SFPE_0': 'pass_energy',
    'Pass Energy': 'pass_energy',
    'Slit Plate': 'slit',
    'Number of Sweeps': 'n_sweeps',
    'Acquisition Mode': 'scan_mode',
    'Region Name': 'scan_region',
    'Instrument': 'instrument',
    'Pressure': 'pressure',
    'User': 'user',
    'Polar': 'theta',
    'Theta': 'theta',
    'Sample': 'sample',
    'Beta': 'beta',
    'Azimuth': 'chi',
    'Location': 'location',
}

rename_standard_attrs = lambda x: rename_dataarray_attrs(x, ATTRS_MAP)
rename_datavar_standard_attrs = lambda x: rename_datavar_attrs(x, ATTRS_MAP)


def walk_scans(path, only_id=False):
    for path, _, files in os.walk(path):
        json_files = [f for f in files if '.json' in f]
        excel_files = [f for f in files if '.xlsx' in f or '.xlx' in f]

        for j in json_files:
            with open(os.path.join(path, j), 'r') as f:
                metadata = json.load(f)

            for scan in metadata:
                if only_id:
                    yield scan['id']
                else:
                    yield scan

        for x in excel_files:
            if 'cleaned' in x or 'cleaned' in path:
                continue

            ds = clean_xlsx_dataset(os.path.join(path, x))
            for file, scan in ds.iterrows():
                scan['file'] = scan.get('path', file)
                scan['short_file'] = file
                if only_id:
                    yield scan['id']
                else:
                    yield scan


_wrappable = {'note', 'data_preparation', 'provenance', 'corrections', 'symmetry_points'}
WHITELIST_KEYS = {'scan_region', 'sample', 'scan_mode', 'id', 'scan_mode'}
FREEZE_PROPS = {'spectrum_type'}


def wrap_attrs_dict(attrs: dict, original_data: DataType = None) -> dict:
    freeze_extra = []
    attrs_copy = copy.deepcopy(attrs)
    for key in _wrappable:
        if key not in attrs_copy:
            continue

        attrs_copy[key] = json.dumps(attrs_copy[key])

    for prop in FREEZE_PROPS:
        if prop not in original_data.attrs:
            try:
                resolved = normalize_to_spectrum(original_data).S
                attrs_copy[prop] = getattr(resolved, prop)
            except AttributeError:
                warnings.warn('Unresolvable attribute: {}'.format(prop))

    for k, v in attrs_copy.items():
        if v is None:
            freeze_extra.append(k)
            attrs_copy[k] = json.dumps(v)
        elif isinstance(v, bool):
            attrs_copy[k] = 1 if v else 0
        elif isinstance(v, (pd.Timestamp, datetime.time,)):
            attrs_copy[k] = v.isoformat()
        elif isinstance(v, (datetime.datetime,)):
            attrs_copy[k] = str(v)
        elif not isinstance(v, (str, float, int,)):
            print('Be careful about type: {}'.format(type(v)))

    def clean_key(key: str):
        return key.replace('#', 'num_')

    attrs_copy = {k: v for k, v in attrs_copy.items() if len(k)}
    attrs_copy = {clean_key(k): v for k, v in attrs_copy.items()}

    attrs_copy['freeze_extra'] = json.dumps(freeze_extra)
    return attrs_copy


def unwrap_attrs_dict(attrs: dict) -> dict:
    attrs_copy = copy.deepcopy(attrs)
    freeze_extra = json.loads(attrs_copy.pop('freeze_extra', '[]'))

    for key in _wrappable:
        if key not in attrs_copy:
            continue

        try:
            attrs_copy[key] = json.loads(attrs_copy[key])
        except json.JSONDecodeError:
            pass

    for frozen_extra in freeze_extra:
        try:
            attrs_copy[frozen_extra] = json.loads(attrs_copy[frozen_extra])
        except json.JSONDecodeError:
            warnings.warn('Was unable to unfreeze attribute "{}"'.format(frozen_extra))

    return attrs_copy


wrap_datavar_attrs = lift_datavar_attrs(wrap_attrs_dict)
unwrap_datavar_attrs = lift_datavar_attrs(unwrap_attrs_dict)
wrap_dataarray_attrs = lift_dataarray_attrs(wrap_attrs_dict)
unwrap_dataarray_attrs = lift_dataarray_attrs(unwrap_attrs_dict)



import numpy as np
from matplotlib import pyplot as plt
from matplotlib import animation

import arpes.config
import xarray as xr
from arpes.plotting.utils import path_for_plot
from arpes.provenance import save_plot_provenance

__all__ = ('plot_movie',)


@save_plot_provenance
def plot_movie(data: xr.DataArray, time_dim, interval=None,
               fig=None, ax=None, out=None, **kwargs):
    if not isinstance(data, xr.DataArray):
        raise TypeError('You must provide a DataArray')

    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 7))

    cmap = arpes.config.SETTINGS.get('interactive', {}).get('palette', 'viridis')
    vmax = data.max().item()
    vmin = data.min().item()

    if data.S.is_subtracted:
        cmap = 'RdBu'
        vmax = np.max([np.abs(vmin), np.abs(vmax)])
        vmin = -vmax

    if 'vmax' in kwargs:
        vmax = kwargs.pop('vmax')
    if 'vmin' in kwargs:
        vmin = kwargs.pop('vmin')

    plot = data.mean(time_dim).transpose().plot(vmax=vmax, vmin=vmin, cmap=cmap, **kwargs)

    def init():
        plot.set_array(np.asarray([]))
        return plot,

    animation_coords = data.coords[time_dim].values

    def animate(i):
        coordinate = animation_coords[i]
        data_for_plot = data.sel(**dict([[time_dim, coordinate]]))
        plot.set_array(data_for_plot.values.T.ravel())
        return plot,

    if interval:
        computed_interval = interval
    else:
        computed_interval = 100

    anim = animation.FuncAnimation(
        fig, animate, init_func=init,
        repeat=500,
        frames=len(animation_coords), interval=computed_interval, blit=True)

    Writer = animation.writers['ffmpeg']
    writer = Writer(fps=1000 / computed_interval, metadata=dict(artist='Me'), bitrate=1800)

    if out is not None:
        anim.save(path_for_plot(out), writer=writer)
        return path_for_plot(out)

    #plt.show()
    return anim



import warnings

import h5py
import numpy as np

import xarray as xr
from arpes.endstations import (HemisphericalEndstation, SingleFileEndstation,
                               SynchrotronEndstation)
from arpes.endstations.nexus_utils import read_data_attributes_from
from arpes.preparation import disambiguate_coordinates

__all__ = ('ANTARESEndstation',)


mono = [['ANTARES', 'Monochromator'],
        ['exitSlitAperture', 'resolution', 'currentGratingName', 'currentSlotName', 'energy']]
user_info = [['User'],
             ['email', 'address', 'affiliation', 'name', 'telephone_number']]
misc = [[], ['comment_conditions', 'experimental_frame', 'start_time']]


general_paths = [mono, user_info, misc]

# not generally needed
mbs_general_location = [[], ['PI-X', 'PI-Y', 'PI-Z', 'Phi', 'Theta']]
mbs_general_paths = [mbs_general_location]

# To be run inside ANTARES/MBSAcquisition_{idx}
mbs_acquisition_spectrometer = [[], ['Frames', 'LensMode', 'PASSENERGY', 'DeflX', 'DeflY', 'CenterKE', 'StepSize',
                                     'StartX', 'StartY', 'EndX', 'EndY', 'StartKE', 'NoSlices', 'NoScans']]
mbs_acquisition_paths = [mbs_acquisition_spectrometer]


def parse_axis_name_from_long_name(name):
    return name.split('/')[-1].replace("'", '')


def infer_scan_type_from_data(group):
    """
    Because ANTARES stores every possible data type in the NeXuS file format, zeroing information that is
    not used, we have to determine which data folder to use on the basis of what kind of scan was done.
    """
    scan_name = str(group['scan_config']['name'].value)

    if 'DeflX' in scan_name:
        # Fermi Surface, might need to be more robust
        return 'data_09'

    if 'Scan2D_MBS' in scan_name:
        # two piezo or two DOF image scan
        return 'data_12'

    raise NotImplementedError(scan_name)


class ANTARESEndstation(HemisphericalEndstation, SynchrotronEndstation, SingleFileEndstation):
    """
    Implements loading text files from the MB Scientific text file format.

    There's not too much metadata here except what comes with the analyzer settings.
    """

    PRINCIPAL_NAME = 'ANTARES'
    ALIASES = []

    RENAME_KEYS = {
        'DeflX': 'psi',
        'deflx': 'psi',
        'energy': 'hv',
        'PASSENERGY': 'pass_energy',
        'LensMode': 'lens_mode',
    }

    def load_top_level_scan(self, group, scan_desc: dict=None, spectrum_index=None):
        dr = self.read_scan_data(group)
        attrs = read_data_attributes_from(group, general_paths)

        try:
            mbs_key = [k for k in list(group['ANTARES'].keys()) if 'MBSAcquisition' in k][0]
            attrs.update(read_data_attributes_from(group['ANTARES'][mbs_key], mbs_acquisition_paths))
        except IndexError:
            pass

        dr = dr.assign_attrs(attrs)
        return xr.Dataset(dict([['spectrum-{}'.format(spectrum_index), dr]]))

    def get_coords(self, group, scan_name, shape):
        # This will have to be modified for data which lacks either a phi or energy axis
        # We will cross this bridge once we have any idea what shape the bridge is in

        dims = list(shape)
        data = group['scan_data']

        # handle actuators
        relaxed_shape = list(shape)
        actuator_list = [k for k in list(data.keys()) if 'actuator' in k]
        actuator_names = [parse_axis_name_from_long_name(
            str(data[act].attrs['long_name'])) for act in actuator_list]
        actuator_list = [data[act][:] for act in actuator_list]

        actuator_dim_order = []
        for act in actuator_list:
            found = relaxed_shape.index(act.shape[-1])
            actuator_dim_order.append(found)
            relaxed_shape[found] = None

        coords = {}

        def take_last(vs):
            while len(vs.shape) > 1:
                vs = vs[0]

            return vs

        for dim_order, name, values in zip(actuator_dim_order, actuator_names, actuator_list):
            name = self.RENAME_KEYS.get(name, name)
            dims[dim_order] = name
            coords[name] = take_last(values)

        # handle standard spectrometer axes, keeping in mind things get stored
        # in different places sometimes for no reasons
        energy_keys = {
            'data_09': ('data_01', 'data_03', 'data_02',),
            'data_12': ('data_04', 'data_06', 'data_05',),
        }
        angle_keys = {
            'data_09': ('data_04', 'data_06', 'data_05',),
            'data_12': ('data_07', 'data_09', 'data_08',),
        }

        e_keys = energy_keys[scan_name]
        ang_keys = angle_keys[scan_name]
        energy = data[e_keys[0]][0], data[e_keys[1]][0], data[e_keys[2]][0]
        angle = data[ang_keys[0]][0], data[ang_keys[1]][0], data[ang_keys[2]][0]

        def get_first(item):
            if isinstance(item, np.ndarray):
                return item.ravel()[0]

            return item

        def build_axis(low, high, step_size):
            # this might not work out to be the right thing to do, we will see
            low, high, step_size = get_first(low), get_first(high), get_first(step_size)
            est_n = int((high - low) / step_size)

            closest = None
            diff = np.inf
            idx = None
            for i, s in enumerate(shape):
                if closest is None or np.abs(s - est_n) < diff:
                    idx = i
                    diff = np.abs(s - est_n)
                    closest = s

            if diff is not 0:
                warnings.warn('Could not identify axis by length.')
            return np.linspace(low, high, closest, endpoint=False), idx

        energy, energy_idx = build_axis(*energy)
        angle, angle_idx = build_axis(*angle)

        dims[energy_idx] = 'eV'
        dims[angle_idx] = 'phi'
        coords['eV'] = energy
        coords['phi'] = angle * np.pi / 180

        return dims, coords

    def read_scan_data(self, group):
        """
        Reads the scan data stored in /scan_data/data_{idx} as appropriate for the type of file.
        """
        data_key = infer_scan_type_from_data(group)
        data_group = group['scan_data'][data_key]
        data = data_group[:]

        dims, coords = self.get_coords(group, data_key, shape=data.shape)

        return xr.DataArray(data, coords=coords, dims=dims)

    def load_single_frame(self, frame_path: str=None, scan_desc: dict=None, **kwargs):
        f = h5py.File(frame_path)
        top_level = list(f.keys())

        loaded = [self.load_top_level_scan(f[key], scan_desc, spectrum_index=i) for i, key in enumerate(top_level)]

        if isinstance(loaded, list) and loaded:
            loaded = disambiguate_coordinates(loaded, ['phi', 'eV'])
            loaded = xr.merge(loaded)
        else:
            loaded = loaded[0]
            loaded.rename({'spectrum-1': 'spectrum'})

        loaded = loaded.assign_attrs(**{self.RENAME_KEYS.get(k, k): v for k, v in loaded.attrs.items()})

        return loaded

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict=None):
        def check_attrs(s):
            for k in ['psi', 'hv', 'lens_mode', 'pass_energy']:
                try:
                    if isinstance(s.attrs[k], (np.ndarray, list, tuple,)):
                        s.attrs[k] = s.attrs[k][0]
                    elif isinstance(s.attrs[k], bytes):
                        s.attrs[k] = s.attrs[k].decode()
                except (TypeError, KeyError):
                    pass

        ls = [data] + data.S.spectra
        for l in ls:
            check_attrs(l)

        # TODO fix this
        defaults = {'z': 0, 'x': 0, 'y': 0, 'alpha': 0, 'chi': 0, 'theta': 0, 'beta': 0, 'hv': 100, 'psi': 0}
        for k, v in defaults.items():
            data.attrs[k] = data.attrs.get(k, v)
            for s in data.S.spectra:
                s.attrs[k] = s.attrs.get(k, v)

        return super().postprocess_final(data, scan_desc)



"""
Contains tools pertaining to statistical boostraps. It can sometimes be difficult
to assess when bootstraps are appropriate, so make sure to consider this before you
just stick a bootstrap around your code and stuff the resultant error bar
into your papers.

This is most useful on data coming from ToF experiments, where individual electron
arrivals are counted, but even here you must be aware of tricky aspects of
the experiment: ToF-ARPES analyzers are not perfect, their efficiency can vary dramatically
across the detector due to MCP burn-in, and electron aberration and focusing
must be considered.
"""

import copy
import functools
import random

from typing import Union
import numpy as np
from tqdm import tqdm_notebook

import xarray as xr
from arpes.analysis.sarpes import to_intensity_polarization
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import lift_dataarray_to_generic
from arpes.utilities.normalize import normalize_to_spectrum
from arpes.utilities.region import normalize_region

__all__ = ('bootstrap', 'estimate_prior_adjustment',
           'resample_true_counts', 'bootstrap_counts',
           'bootstrap_intensity_polarization',)


@update_provenance('Estimate prior')
def estimate_prior_adjustment(data: DataType, region: Union[dict, str] = None) -> float:
    r"""
    Estimates the parameters of a distribution generating the intensity
    histogram of pixels in a spectrum. In a perfectly linear, single-electron
    single-count detector, this would be a poisson distribution with
    \lambda=mean(counts) over the window. Despite this, we can estimate \lambda
    phenomenologically and verify that a Poisson distribution provides a good
    prior for the data, allowing us to perform statistical bootstrapping.

    You should use this with a spectrum that has uniform intensity, i.e. with a
    copper reference or similar.

    :param data:
    :return: returns sigma / mu, adjustment factor for the Poisson distribution
    """
    data = normalize_to_spectrum(data)

    if region is None:
        region = 'copper_prior'

    region = normalize_region(region)

    if 'cycle' in data.dims:
        data = data.sum('cycle')

    data = data.S.zero_spectrometer_edges().S.region_sel(region)
    values = data.values.ravel()
    values = values[np.where(values)]
    return np.std(values) / np.mean(values)


@update_provenance('Resample cycle dimension')
@lift_dataarray_to_generic
def resample_cycle(data: xr.DataArray, **kwargs):
    """
    Perform a non-parametric bootstrap using a cycle coordinate for statistically independent observations.
    :param data:
    :param kwargs:
    :return:
    """

    n_cycles = len(data.cycle)
    which = [random.randint(0, n_cycles - 1) for _ in range(n_cycles)]

    resampled = data.isel(cycle=which).sum('cycle', keep_attrs=True)

    if 'id' in resampled.attrs:
        del resampled.attrs['id']

    return resampled


@update_provenance('Resample with prior adjustment')
@lift_dataarray_to_generic
def resample(data: xr.DataArray, prior_adjustment=1, **kwargs):
    resampled = xr.DataArray(
        np.random.poisson(lam=data.values * prior_adjustment, size=data.values.shape),
        coords=data.coords,
        dims=data.dims,
        attrs=data.attrs
    )

    if 'id' in resampled.attrs:
        del resampled.attrs['id']

    return resampled


@update_provenance('Resample electron-counted data')
@lift_dataarray_to_generic
def resample_true_counts(data: xr.DataArray) -> xr.DataArray:
    """
    Resamples histogrammed data where each count represents an actual electron.
    :param data:
    :return:
    """

    resampled = xr.DataArray(
        np.random.poisson(lam=data.values, size=data.values.shape),
        coords=data.coords,
        dims=data.dims,
        attrs=data.attrs,
    )

    if 'id' in resampled.attrs:
        del resampled.attrs['id']

    return resampled


@update_provenance('Bootstrap true electron counts')
@lift_dataarray_to_generic
def bootstrap_counts(data: DataType, N=1000, name=None) -> xr.Dataset:
    """
    Parametric bootstrap for the number of counts in each detector channel for a
    time of flight/DLD detector, where each count represents an actual particle.

    This function also introspects the data passed to determine whether there is a
    spin degree of freedom, and will bootstrap appropriately.

    Currently we build all the samples at once instead of using a rolling algorithm.
    :param data:
    :return:
    """

    assert (data.name is not None or name is not None)
    name = data.name if data.name is not None else name

    desc_fragment = ' {}'.format(name)

    resampled_sets = []
    for _ in tqdm_notebook(range(N), desc='Resampling{}...'.format(desc_fragment)):
        resampled_sets.append(resample_true_counts(data))

    resampled_arr = np.stack([s.values for s in resampled_sets], axis=0)
    std = np.std(resampled_arr, axis=0)
    std = xr.DataArray(std, data.coords, data.dims)
    mean = np.mean(resampled_arr, axis=0)
    mean = xr.DataArray(mean, data.coords, data.dims)

    data_vars = {}
    data_vars[name] = mean
    data_vars[name + '_std'] = std

    return xr.Dataset(data_vars=data_vars, coords=data.coords, attrs=data.attrs.copy())


@update_provenance('Bootstrap spin detector polarization and intensity')
def bootstrap_intensity_polarization(data, N=100):
    """
    Uses the parametric bootstrap to get uncertainties on the intensity and polarization of ToF-SARPES data.
    :param data:
    :return:
    """

    bootstrapped_polarization = bootstrap(to_intensity_polarization)
    return bootstrapped_polarization(data, N=N)


def bootstrap(fn, skip=None, resample_method=None):
    if skip is None:
        skip = []

    skip = set(skip)

    if resample_method is None:
        resample_fn = resample
    elif resample_method == 'cycle':
        resample_fn = resample_cycle

    def bootstrapped(*args, N=20, prior_adjustment=1, **kwargs):
        # examine args to determine which to resample
        resample_indices = [i for i, arg in enumerate(args) if isinstance(arg, (xr.DataArray, xr.Dataset)) and i not in skip]
        data_is_arraylike = False

        runs = []

        def get_label(i):
            if isinstance(args[i], xr.Dataset):
                return 'xr.Dataset: [{}]'.format(', '.join(args[i].data_vars.keys()))
            if args[i].name:
                return args[i].name

            try:
                return args[i].attrs['id']
            except KeyError:
                return 'Label-less DataArray'

        print('Resampling args: {}'.format(','.join([get_label(i) for i in resample_indices])))

        # examine kwargs to determine which to resample
        resample_kwargs = [k for k, v in kwargs.items() if isinstance(v, xr.DataArray) and k not in skip]
        print('Resampling kwargs: {}'.format(','.join(resample_kwargs)))

        print('Fair warning 1: Make sure you understand whether it is appropriate to resample your data.')
        print('Fair warning 2: Ensure that the data to resample is in a DataArray and not a Dataset')

        for _ in tqdm_notebook(range(N), desc='Resampling...'):
            new_args = list(args)
            new_kwargs = copy.copy(kwargs)
            for i in resample_indices:
                new_args[i] = resample_fn(args[i], prior_adjustment=prior_adjustment)
            for k in resample_kwargs:
                new_kwargs[k] = resample_fn(kwargs[k], prior_adjustment=prior_adjustment)


            run = fn(*new_args, **new_kwargs)
            if isinstance(run, (xr.DataArray, xr.Dataset)):
                data_is_arraylike = True
            runs.append(run)

        if data_is_arraylike:
            for i, run in enumerate(runs):
                run.coords['bootstrap'] = i

            return xr.concat(runs, dim='bootstrap')

        return runs

    return functools.wraps(fn)(bootstrapped)



"""
Implements some pipelines for basic tasks.
"""

import xarray as xr
from arpes.corrections import (apply_photon_energy_fermi_edge_correction,
                               apply_quadratic_fermi_edge_correction)

from .pipeline import compose, pipeline
from .preparation import dim_normalizer
from .provenance import update_provenance
from .utilities import conversion

__all__ = ['convert_scan_to_kspace', 'convert_scan_to_kspace_no_corr',
           'tr_prep_scan_ammeter', 'tr_prep_scan_simple', 'convert_mc_map_to_kspace_fft_filter',
           'prep_for_d2']

@pipeline()
def correct_e_fermi_hv(arr: xr.DataArray):
    if isinstance(arr, xr.Dataset):
        arr = arr.raw

    if 'hv' not in arr.dims:
        return arr

    return apply_photon_energy_fermi_edge_correction(arr)


@pipeline()
def correct_e_fermi_spectrometer(arr: xr.DataArray):
    if 'phi' not in arr.dims:
        return arr

    return apply_quadratic_fermi_edge_correction(arr)


@pipeline()
@update_provenance('Sum cycle dimension')
def sum_cycles(arr: xr.DataArray):
    if 'cycle' not in arr.dims:
        return arr

    return arr.sum('cycle', keep_attrs=True)


# TODO: implement pipelines better so that arguments are interned correctly
# in the case where a pipeline step has extra arguments

# Pipelines should never include data loading
# Scans are already normalized at this point, they should be whenever they are first
# interned in the netCDF format
convert_scan_to_kspace = compose(
    pipeline('normalize_hv_axis')(dim_normalizer('hv')),
    correct_e_fermi_hv,
    correct_e_fermi_spectrometer,
    pipeline()(conversion.convert_to_kspace),
)

convert_scan_to_kspace_no_corr = compose(
    pipeline('normalize_hv_axis')(dim_normalizer('hv')),
    pipeline()(conversion.convert_to_kspace),
)

prep_for_d2 = compose(
    pipeline('normalize_hv_axis')(dim_normalizer('hv')),
    correct_e_fermi_hv,
    correct_e_fermi_spectrometer,
)

@pipeline()
def fft_clean_mc_map(data: xr.DataArray):
    from arpes.analysis.fft import fft_filter

    beta_filter = [
        {'beta': slice(0.65, 1.)},
        {'beta': slice(-1., -0.65)},
    ]
    return fft_filter(data, beta_filter)


convert_mc_map_to_kspace_fft_filter = compose(
    pipeline('normalize_polar_axis')(dim_normalizer('beta')),
    fft_clean_mc_map,
    pipeline()(conversion.convert_to_kspace),
)

tr_prep_scan_simple = compose(
    pipeline('normalize_cycle_axis')(dim_normalizer('cycle')),
    sum_cycles,
)

tr_prep_scan_ammeter = compose(
    #normalize_from_ammeter,
    sum_cycles,
)
